# Copyright (c) 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
from unittest import mock

import six

from barbican.model import models
from barbican.model import repositories
from barbican.queue import server
from barbican.tasks import common
from barbican.tests import database_utils
from barbican.tests import utils


class WhenUsingTransactionalDecorator(utils.BaseTestCase):
    """Test using the 'transactional' decorator in server.py.

    Note that only the 'I am a server' logic is tested here, as the alternate
    mode is only used for direct invocation of Task methods in the standalone
    server mode, which is also thoroughly tested in WhenUsingBeginTypeOrderTask
    below.
    """

    def setUp(self):
        super(WhenUsingTransactionalDecorator, self).setUp()

        # Ensure we always thing we are in 'I am a server' mode.
        is_server_side_config = {
            'return_value': True
        }
        self.is_server_side_patcher = mock.patch(
            'barbican.queue.is_server_side',
            **is_server_side_config
        )
        self.is_server_side_patcher.start()

        self.commit_patcher = mock.patch(
            'barbican.model.repositories.commit'
        )
        self.commit_mock = self.commit_patcher.start()

        self.rollback_patcher = mock.patch(
            'barbican.model.repositories.rollback'
        )
        self.rollback_mock = self.rollback_patcher.start()

        self.clear_patcher = mock.patch(
            'barbican.model.repositories.clear'
        )
        self.clear_mock = self.clear_patcher.start()

        self.args = ('foo', 'bar')
        self.kwargs = {'k_foo': 1, 'k_bar': 2}

        # Class/decorator under test.
        class TestClass(object):
            my_args = None
            my_kwargs = None
            is_exception_needed = False

            @server.transactional
            def test_method(self, *args, **kwargs):
                if self.is_exception_needed:
                    raise ValueError()
                self.my_args = args
                self.my_kwargs = kwargs
        self.test_object = TestClass()

    def tearDown(self):
        super(WhenUsingTransactionalDecorator, self).tearDown()
        self.is_server_side_patcher.stop()
        self.commit_patcher.stop()
        self.rollback_patcher.stop()
        self.clear_patcher.stop()

    def test_should_commit(self):
        self.test_object.test_method(*self.args, **self.kwargs)

        self.assertEqual(self.args, self.test_object.my_args)
        self.assertEqual(self.kwargs, self.test_object.my_kwargs)

        self.assertEqual(1, self.commit_mock.call_count)
        self.assertEqual(0, self.rollback_mock.call_count)
        self.assertEqual(1, self.clear_mock.call_count)

    def test_should_rollback(self):
        self.test_object.is_exception_needed = True

        self.test_object.test_method(*self.args, **self.kwargs)

        self.assertEqual(0, self.commit_mock.call_count)
        self.assertEqual(1, self.rollback_mock.call_count)
        self.assertEqual(1, self.clear_mock.call_count)


class WhenUsingRetryableOrderDecorator(utils.BaseTestCase):
    """Test using the 'retryable_order' decorator in server.py."""

    def setUp(self):
        super(WhenUsingRetryableOrderDecorator, self).setUp()

        self.schedule_retry_tasks_patcher = mock.patch(
            'barbican.queue.server.schedule_order_retry_tasks'
        )
        self.schedule_retry_tasks_mock = (
            self.schedule_retry_tasks_patcher.start()
        )

        self.order_id = 'order-id'
        self.args = ('foo', 'bar')
        self.kwargs = {'k_foo': 1, 'k_bar': 2}

        # Class/decorator under test.
        class TestClass(object):
            self.order_id = None
            my_args = None
            my_kwargs = None
            is_exception_needed = False
            result = common.FollowOnProcessingStatusDTO()

            @server.retryable_order
            def test_method(self, order_id, *args, **kwargs):
                if self.is_exception_needed:
                    raise ValueError()
                self.order_id = order_id
                self.my_args = args
                self.my_kwargs = kwargs
                return self.result
        self.test_object = TestClass()
        self.test_method = TestClass.test_method

    def tearDown(self):
        super(WhenUsingRetryableOrderDecorator, self).tearDown()
        self.schedule_retry_tasks_patcher.stop()

    def test_should_successfully_schedule_a_task_for_retry(self):
        self.test_object.test_method(self.order_id, *self.args, **self.kwargs)

        self.assertEqual(self.order_id, self.test_object.order_id)
        self.assertEqual(self.args, self.test_object.my_args)
        self.assertEqual(self.kwargs, self.test_object.my_kwargs)

        self.assertEqual(1, self.schedule_retry_tasks_mock.call_count)
        self.schedule_retry_tasks_mock.assert_called_with(
            mock.ANY,
            self.test_object.result,
            self.order_id,
            *self.args,
            **self.kwargs)

    def test_retry_should_not_be_scheduled_if_exception_is_raised(self):
        self.test_object.is_exception_needed = True

        self.assertRaises(
            ValueError,
            self.test_object.test_method,
            self.order_id,
            self.args,
            self.kwargs,
        )

        self.assertEqual(0, self.schedule_retry_tasks_mock.call_count)


class WhenCallingScheduleOrderRetryTasks(database_utils.RepositoryTestCase):
    """Test calling schedule_order_retry_tasks() in server.py."""

    def setUp(self):
        super(WhenCallingScheduleOrderRetryTasks, self).setUp()

        self.project = database_utils.create_project()
        self.order = database_utils.create_order(self.project)
        database_utils.get_session().commit()

        self.repo = repositories.OrderRetryTaskRepo()

        self.result = common.FollowOnProcessingStatusDTO()

        self.args = ['args-foo', 'args-bar']
        self.kwargs = {'order_id': self.order.id, 'foo': 1, 'bar': 2}
        self.date_to_retry_at = (
            datetime.datetime.utcnow() + datetime.timedelta(
                milliseconds=self.result.retry_msec)
        )

    def test_should_not_schedule_task_due_to_no_result(self):
        retry_rpc_method = server.schedule_order_retry_tasks(None, None, None)

        self.assertIsNone(retry_rpc_method)

    def test_should_not_schedule_task_due_to_no_action_required_result(self):
        self.result.retry_task = common.RetryTasks.NO_ACTION_REQUIRED

        retry_rpc_method = server.schedule_order_retry_tasks(
            None, self.result, None)

        self.assertIsNone(retry_rpc_method)

    def test_should_schedule_invoking_task_for_retry(self):
        self.result.retry_task = common.RetryTasks.INVOKE_SAME_TASK

        # Schedule this test method as the passed-in 'retry' function.
        retry_rpc_method = server.schedule_order_retry_tasks(
            self.test_should_schedule_invoking_task_for_retry,
            self.result,
            None,  # Not used.
            *self.args,
            **self.kwargs)
        database_utils.get_session().commit()  # Flush to the database.

        self.assertEqual(
            'test_should_schedule_invoking_task_for_retry', retry_rpc_method)

    def test_should_schedule_certificate_status_task_for_retry(self):
        self.result.retry_task = (
            common.RetryTasks.INVOKE_CERT_STATUS_CHECK_TASK
        )

        # Schedule this test method as the passed-in 'retry' function.
        retry_rpc_method = server.schedule_order_retry_tasks(
            None,  # Should be ignored for non-self retries.
            self.result,
            None,  # Not used.
            *self.args,
            **self.kwargs)
        database_utils.get_session().commit()  # Flush to the database.

        self.assertEqual(
            'check_certificate_status', retry_rpc_method)
        self._verify_retry_task_entity(
            'check_certificate_status')

    def _verify_retry_task_entity(self, retry_task):
        # Retrieve the task retry entity created above and verify it.
        entities, offset, limit, total = self.repo.get_by_create_date()
        self.assertEqual(1, total)
        retry_model = entities[0]
        self.assertEqual(retry_task, retry_model.retry_task)
        self.assertEqual(self.args, retry_model.retry_args)
        self.assertEqual(self.kwargs, retry_model.retry_kwargs)
        self.assertEqual(0, retry_model.retry_count)

        # Compare retry_at times.
        # Note that the expected retry_at time is computed at setUp() time, but
        # the retry_at time on the task retry entity/model is computed and set
        # a few milliseconds after this setUp() time, hence they will vary by a
        # small amount of time.
        delta = retry_model.retry_at - self.date_to_retry_at
        delta_seconds = delta.seconds
        self.assertLessEqual(delta_seconds, 2)


class WhenCallingTasksMethod(utils.BaseTestCase):
    """Test calling methods on the Tasks class."""

    def setUp(self):
        super(WhenCallingTasksMethod, self).setUp()

        # Mock the 'am I a server process?' flag used by the decorator around
        #   all task methods. Since this test class focuses on testing task
        #   method behaviors, this flag is set to false to allow for direct
        #   testing of these tasks without database transactional interference.
        is_server_side_config = {
            'return_value': False
        }
        self.is_server_side_patcher = mock.patch(
            'barbican.queue.is_server_side',
            **is_server_side_config
        )
        self.is_server_side_patcher.start()

        self.tasks = server.Tasks()

    def tearDown(self):
        super(WhenCallingTasksMethod, self).tearDown()
        self.is_server_side_patcher.stop()

    @mock.patch('barbican.queue.server.schedule_order_retry_tasks')
    @mock.patch('barbican.tasks.resources.BeginTypeOrder')
    def test_should_process_begin_order(self, mock_begin_order, mock_schedule):
        method = mock_begin_order.return_value.process_and_suppress_exceptions
        method.return_value = 'result'

        self.tasks.process_type_order(
            None, self.order_id, self.external_project_id, self.request_id)

        mock_process = mock_begin_order.return_value
        mock_process.process_and_suppress_exceptions.assert_called_with(
            self.order_id, self.external_project_id)
        mock_schedule.assert_called_with(
            mock.ANY, 'result', None, 'order1234',
            'keystone1234', 'request1234')

    @mock.patch('barbican.queue.server.schedule_order_retry_tasks')
    @mock.patch('barbican.tasks.resources.CheckCertificateStatusOrder')
    def test_should_check_certificate_order(
            self, mock_check_cert, mock_schedule):
        method = mock_check_cert.return_value.process_and_suppress_exceptions
        method.return_value = 'result'

        self.tasks.check_certificate_status(
            None, self.order_id, self.external_project_id, self.request_id)

        mock_process = mock_check_cert.return_value
        mock_process.process_and_suppress_exceptions.assert_called_with(
            self.order_id, self.external_project_id
        )
        mock_schedule.assert_called_with(
            mock.ANY, 'result', None, 'order1234',
            'keystone1234', 'request1234')

    @mock.patch('barbican.tasks.resources.BeginTypeOrder')
    def test_process_order_catch_exception(self, mock_begin_order):
        """Test that BeginTypeOrder's process() handles all exceptions."""
        mock_begin_order.return_value._process.side_effect = Exception()

        self.tasks.process_type_order(None, self.order_id,
                                      self.external_project_id,
                                      self.request_id)


class WhenUsingTaskServer(database_utils.RepositoryTestCase):
    """Test using the asynchronous task client.

    This test suite performs a full-stack test of worker-side task
    processing (except for queue interactions, which are mocked). This
    includes testing database commit and session close behaviors.
    """

    def setUp(self):
        super(WhenUsingTaskServer, self).setUp()

        # Queue target mocking setup.
        self.target = 'a target value here'
        queue_get_target_config = {
            'return_value': self.target
        }
        self.queue_get_target_patcher = mock.patch(
            'barbican.queue.get_target',
            **queue_get_target_config
        )
        self.queue_get_target_mock = self.queue_get_target_patcher.start()

        # Queue server mocking setup.
        self.server_mock = mock.MagicMock()
        self.server_mock.start.return_value = None
        self.server_mock.stop.return_value = None
        queue_get_server_config = {
            'return_value': self.server_mock
        }
        self.queue_get_server_patcher = mock.patch(
            'barbican.queue.get_server',
            **queue_get_server_config
        )
        self.queue_get_server_mock = self.queue_get_server_patcher.start()

        self.server = server.TaskServer()

        # Add an order to the in-memory database.
        self.external_id = 'keystone-id'
        project = database_utils.create_project(
            external_id=self.external_id)
        self.order = database_utils.create_order(
            project=project)
        self.request_id = 'request1234'

    def tearDown(self):
        super(WhenUsingTaskServer, self).tearDown()
        self.queue_get_target_patcher.stop()
        self.queue_get_server_patcher.stop()

    def test_should_start(self):
        self.server.start()

        self.queue_get_target_mock.assert_called_with()
        self.queue_get_server_mock.assert_called_with(
            target=self.target, endpoints=[self.server])
        self.server_mock.start.assert_called_with()

    def test_should_stop(self):
        self.server.stop()
        self.queue_get_target_mock.assert_called_with()
        self.queue_get_server_mock.assert_called_with(
            target=self.target, endpoints=[self.server])
        self.server_mock.stop.assert_called_with()

    def test_process_bogus_begin_type_order_should_not_rollback(self):
        order_id = self.order.id
        self.order.type = 'bogus-type'  # Force error out of business logic.

        # Invoke process, including the transactional decorator that terminates
        # the session when it is done. Hence we must re-retrieve the order for
        # verification afterwards.
        self.server.process_type_order(
            None, self.order.id, self.external_id, self.request_id)

        order_repo = repositories.get_order_repository()
        order_result = order_repo.get(order_id, self.external_id)

        self.assertEqual(models.States.ERROR, order_result.status)
        self.assertEqual(
            six.u(
                'Process TypeOrder failure seen - '
                'please contact site administrator.'),
            order_result.error_reason)
        self.assertEqual(
            six.u('500'),
            order_result.error_status_code)



# Copyright (c) 2015 Cisco Systems
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from barbican.tests import utils
from functionaltests.api import base
from functionaltests.api.v1.behaviors import acl_behaviors
from functionaltests.api.v1.behaviors import container_behaviors
from functionaltests.api.v1.behaviors import secret_behaviors
from functionaltests.api.v1.models import acl_models
from functionaltests.api.v1.models import container_models
from functionaltests.api.v1.models import secret_models
from functionaltests.common import config


CONF = config.get_config()
admin_a = CONF.rbac_users.admin_a
creator_a = CONF.rbac_users.creator_a
observer_a = CONF.rbac_users.observer_a
auditor_a = CONF.rbac_users.auditor_a
admin_b = CONF.rbac_users.admin_b
observer_b = CONF.rbac_users.observer_b


def get_acl_default():
    return {'read': {'project-access': True}}


def get_acl_one():
    return {'read': {'users': ['reader1'], 'project-access': False}}


def get_acl_two():
    return {'read': {'users': ['reader2'], 'project-access': False}}


test_data_set_secret_acl = {
    'with_admin_a': {'user': admin_a, 'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'expected_return': 403},
    'with_admin_b': {'user': admin_b, 'expected_return': 403},
    'with_observer_b': {'user': observer_b, 'expected_return': 403},
}

test_data_get_secret_acl = {
    'with_admin_a': {'user': admin_a, 'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'expected_return': 200},
    'with_auditor_a': {'user': auditor_a, 'expected_return': 403},
    'with_admin_b': {'user': admin_b, 'expected_return': 403},
    'with_observer_b': {'user': observer_b, 'expected_return': 403},
}

test_data_update_secret_acl = {
    'with_admin_a': {'user': admin_a, 'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'expected_return': 403},
    'with_admin_b': {'user': admin_b, 'expected_return': 403},
    'with_observer_b': {'user': observer_b, 'expected_return': 403},
}

test_data_delete_secret_acl = {
    'with_admin_a': {'user': admin_a, 'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'expected_return': 403},
    'with_admin_b': {'user': admin_b, 'expected_return': 403},
    'with_observer_b': {'user': observer_b, 'expected_return': 403},
}

test_data_set_container_acl = {
    'with_admin_a': {'user': admin_a, 'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'expected_return': 403},
    'with_admin_b': {'user': admin_b, 'expected_return': 403},
    'with_observer_b': {'user': observer_b, 'expected_return': 403},
}

test_data_get_container_acl = {
    'with_admin_a': {'user': admin_a, 'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'expected_return': 200},
    'with_auditor_a': {'user': auditor_a, 'expected_return': 403},
    'with_admin_b': {'user': admin_b, 'expected_return': 403},
    'with_observer_b': {'user': observer_b, 'expected_return': 403},
}

test_data_update_container_acl = {
    'with_admin_a': {'user': admin_a, 'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'expected_return': 403},
    'with_admin_b': {'user': admin_b, 'expected_return': 403},
    'with_observer_b': {'user': observer_b, 'expected_return': 403},
}

test_data_delete_container_acl = {
    'with_admin_a': {'user': admin_a, 'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'expected_return': 403},
    'with_admin_b': {'user': admin_b, 'expected_return': 403},
    'with_observer_b': {'user': observer_b, 'expected_return': 403},
}


@utils.parameterized_test_case
class RBACAclsTestCase(base.TestCase):
    """Functional tests exercising RBAC Policies for ACL Operations"""
    def setUp(self):
        super(RBACAclsTestCase, self).setUp()
        self.secret_behaviors = secret_behaviors.SecretBehaviors(self.client)
        self.container_behaviors = container_behaviors.ContainerBehaviors(
            self.client)
        self.acl_behaviors = acl_behaviors.AclBehaviors(self.client)

    def tearDown(self):
        self.acl_behaviors.delete_all_created_acls()
        self.secret_behaviors.delete_all_created_secrets()
        self.container_behaviors.delete_all_created_containers()
        super(RBACAclsTestCase, self).tearDown()

    @utils.parameterized_dataset(test_data_set_secret_acl)
    def test_set_secret_acl(self, user, expected_return):
        secret_ref = self.store_secret()
        status = self.set_secret_acl(secret_ref, get_acl_one(),
                                     user_name=user)
        self.assertEqual(expected_return, status)

    @utils.parameterized_dataset(test_data_get_secret_acl)
    def test_get_secret_acl(self, user, expected_return):
        secret_ref = self.store_secret()
        status = self.set_secret_acl(secret_ref, get_acl_one())
        self.assertEqual(200, status)
        resp = self.acl_behaviors.get_acl(secret_ref + '/acl', user_name=user)
        self.assertEqual(expected_return, resp.status_code)
        if expected_return == 200:
            self.assertIn('reader1', resp.model.read['users'])
        else:
            self.assertIsNone(resp.model)

    @utils.parameterized_dataset(test_data_update_secret_acl)
    def test_update_secret_acl(self, user, expected_return):
        secret_ref = self.store_secret()
        status = self.set_secret_acl(secret_ref, get_acl_one())
        self.assertEqual(200, status)
        status, model = self.update_secret_acl(secret_ref,
                                               get_acl_two(),
                                               user_name=user)
        self.assertEqual(expected_return, status)
        get_resp = self.acl_behaviors.get_acl(secret_ref + '/acl',
                                              user_name=admin_a)
        if expected_return == 200:
            self.assertIsNotNone(model.acl_ref)
            # verify update happened
            self.assertIn('reader2', get_resp.model.read['users'])
        else:
            self.assertIsNone(model)
            # verify no update happened
            self.assertIn('reader1', get_resp.model.read['users'])

    @utils.parameterized_dataset(test_data_delete_secret_acl)
    def test_delete_secret_acl(self, user, expected_return):
        secret_ref = self.store_secret()
        status = self.set_secret_acl(secret_ref, get_acl_one())
        self.assertEqual(200, status)
        resp = self.acl_behaviors.delete_acl(secret_ref + '/acl',
                                             user_name=user)
        self.assertEqual(expected_return, resp.status_code)
        get_resp = self.acl_behaviors.get_acl(secret_ref + '/acl',
                                              user_name=admin_a)
        if expected_return == 200:
            # verify delete happened (return to default ACL)
            self.assertTrue(get_resp.model.read['project-access'])
        else:
            # verify no delete happened
            self.assertIn('reader1', get_resp.model.read['users'])

    @utils.parameterized_dataset(test_data_set_container_acl)
    def test_set_container_acl(self, user, expected_return):
        container_ref = self.store_container()
        status = self.set_container_acl(container_ref, get_acl_one(),
                                        user_name=user)
        self.assertEqual(expected_return, status)

    @utils.parameterized_dataset(test_data_get_container_acl)
    def test_get_container_acl(self, user, expected_return):
        container_ref = self.store_container()
        status = self.set_container_acl(container_ref, get_acl_one())
        self.assertEqual(200, status)
        resp = self.acl_behaviors.get_acl(container_ref + '/acl',
                                          user_name=user)
        self.assertEqual(expected_return, resp.status_code)
        if expected_return == 200:
            self.assertIn('reader1', resp.model.read['users'])
        else:
            self.assertIsNone(resp.model)

    @utils.parameterized_dataset(test_data_update_container_acl)
    def test_update_container_acl(self, user, expected_return):
        container_ref = self.store_container()
        status = self.set_container_acl(container_ref, get_acl_one())
        self.assertEqual(200, status)
        status, model = self.update_container_acl(container_ref,
                                                  get_acl_two(),
                                                  user_name=user)
        self.assertEqual(expected_return, status)
        get_resp = self.acl_behaviors.get_acl(container_ref + '/acl',
                                              user_name=admin_a)
        if expected_return == 200:
            self.assertIsNotNone(model.acl_ref)
            # verify update happened
            self.assertIn('reader2', get_resp.model.read['users'])
        else:
            self.assertIsNone(model)
            # verify no update happened
            self.assertIn('reader1', get_resp.model.read['users'])

    @utils.parameterized_dataset(test_data_delete_container_acl)
    def test_delete_container_acl(self, user, expected_return):
        container_ref = self.store_container()
        status = self.set_container_acl(container_ref, get_acl_one())
        self.assertEqual(200, status)
        resp = self.acl_behaviors.delete_acl(container_ref + '/acl',
                                             user_name=user)
        self.assertEqual(expected_return, resp.status_code)
        get_resp = self.acl_behaviors.get_acl(container_ref + '/acl',
                                              user_name=admin_a)
        if expected_return == 200:
            # verify delete happened (return to default ACL)
            self.assertTrue(get_resp.model.read['project-access'])
        else:
            # verify no delete happened
            self.assertIn('reader1', get_resp.model.read['users'])

# ----------------------- Helper Functions ---------------------------

    def store_secret(self, user_name=creator_a, admin=admin_a):
        test_model = secret_models.SecretModel(
            **get_default_secret_data())
        resp, secret_ref = self.secret_behaviors.create_secret(
            test_model, user_name=user_name, admin=admin)
        self.assertEqual(201, resp.status_code)
        return secret_ref

    def set_secret_acl(self, secret_ref, acl, user_name=creator_a):
        test_model = acl_models.AclModel(**acl)
        resp = self.acl_behaviors.create_acl(
            secret_ref, test_model, user_name=user_name)
        return resp.status_code

    def update_secret_acl(self, secret_ref, acl, user_name=creator_a):
        test_model = acl_models.AclModel(**acl)
        resp = self.acl_behaviors.update_acl(
            secret_ref + '/acl', test_model, user_name=user_name)
        return resp.status_code, resp.model

    def store_container(self, user_name=creator_a, admin=admin_a):
        secret_ref = self.store_secret(user_name=user_name, admin=admin)

        test_model = container_models.ContainerModel(
            **get_container_req(secret_ref))
        resp, container_ref = self.container_behaviors.create_container(
            test_model, user_name=user_name, admin=admin)
        self.assertEqual(201, resp.status_code)
        return container_ref

    def set_container_acl(self, container_ref, acl, user_name=creator_a):
        test_model = acl_models.AclModel(**acl)
        resp = self.acl_behaviors.create_acl(
            container_ref, test_model, user_name=user_name)
        return resp.status_code

    def update_container_acl(self, container_ref, acl, user_name=creator_a):
        test_model = acl_models.AclModel(**acl)
        resp = self.acl_behaviors.update_acl(
            container_ref + '/acl', test_model, user_name=user_name)
        return resp.status_code, resp.model

# ----------------------- Support Functions ---------------------------


def get_default_secret_data():
    return {
        "name": "AES key",
        "expiration": "2050-02-28T19:14:44.180394",
        "algorithm": "aes",
        "bit_length": 256,
        "mode": "cbc",
        "payload": get_default_payload(),
        "payload_content_type": "application/octet-stream",
        "payload_content_encoding": "base64",
    }


def get_default_payload():
    return 'Z0Y2K2xMb0Yzb2hBOWFQUnB0KzZiUT09'


def get_container_req(secret_ref):
    return {"name": "testcontainer",
            "type": "generic",
            "secret_refs": [{'name': 'secret1', 'secret_ref': secret_ref}]}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from OpenSSL import crypto
from oslo_serialization import base64
import six

from barbican import i18n as u  # noqa
from barbican.plugin.interface import secret_store as s
from barbican.plugin.util import mime_types


def normalize_before_encryption(unencrypted, content_type, content_encoding,
                                secret_type, enforce_text_only=False):
    """Normalize unencrypted prior to plugin encryption processing.

    This normalizes the secrets before they are handed off to the SecretStore
    for storage. This converts all data to Base64 data. If the data is plain
    text then it encoded using utf-8 first and then Base64 encoded. Binary
    data is simply converted to Base64.

    :param str unencrypted: Raw payload
    :param str content_type: The media type for the payload
    :param str content_encoding: Transfer encoding
    :param str secret_type: The type of secret
    :param bool enforce_text_only: Require text content_type or base64
        content_encoding
    :returns: Tuple containing the normalized (base64 encoded) payload and
        the normalized media type.
    """
    if not unencrypted:
        raise s.SecretNoPayloadProvidedException()

    # Validate and normalize content-type.
    normalized_media_type = normalize_content_type(content_type)

    # Process plain-text type.
    if normalized_media_type in mime_types.PLAIN_TEXT:
        # normalize text to binary and then base64 encode it
        b64payload = base64.encode_as_bytes(unencrypted)

    # Process binary type.
    else:
        if not content_encoding:
            b64payload = base64.encode_as_bytes(unencrypted)
        elif content_encoding.lower() == 'base64':
            if not isinstance(unencrypted, six.binary_type):
                b64payload = unencrypted.encode('utf-8')
            else:
                b64payload = unencrypted
        elif enforce_text_only:
            # For text-based protocols (such as the one-step secret POST),
            #   only 'base64' encoding is possible/supported.
            raise s.SecretContentEncodingMustBeBase64()
        else:
            # Unsupported content-encoding request.
            raise s.SecretContentEncodingNotSupportedException(
                content_encoding
            )

    return b64payload, normalized_media_type


def normalize_content_type(content_type):
    """Normalize the content type and validate that it is supported."""
    normalized_mime = mime_types.normalize_content_type(content_type)
    if not mime_types.is_supported(normalized_mime):
        raise s.SecretContentTypeNotSupportedException(content_type)
    return normalized_mime


def analyze_before_decryption(content_type):
    """Determine support for desired content type."""
    if not mime_types.is_supported(content_type):
        raise s.SecretAcceptNotSupportedException(content_type)


def denormalize_after_decryption(unencrypted, content_type):
    """Translate the decrypted data into the desired content type.

    This is called when the raw keys are requested by the user. The secret
    returned from the SecretStore is the unencrypted parameter. This
    'denormalizes' the data back to its binary format.
    """

    # Process plain-text type.
    if content_type in mime_types.PLAIN_TEXT:
        # normalize text to binary string
        try:
            unencrypted = base64.decode_as_text(unencrypted)
        except UnicodeDecodeError:
            raise s.SecretAcceptNotSupportedException(content_type)

    # Process binary type.
    elif content_type in mime_types.BINARY:
        unencrypted = base64.decode_as_bytes(unencrypted)
    else:
        raise s.SecretContentTypeNotSupportedException(content_type)

    return unencrypted


def convert_pem_to_der(pem, secret_type):
    if secret_type == s.SecretType.PRIVATE:
        return _convert_private_pem_to_der(pem)
    elif secret_type == s.SecretType.PUBLIC:
        return _convert_public_pem_to_der(pem)
    elif secret_type == s.SecretType.CERTIFICATE:
        return _convert_certificate_pem_to_der(pem)
    else:
        reason = u._("Secret type can not be converted to DER")
        raise s.SecretGeneralException(reason=reason)


def convert_der_to_pem(der, secret_type):
    if secret_type == s.SecretType.PRIVATE:
        return _convert_private_der_to_pem(der)
    elif secret_type == s.SecretType.PUBLIC:
        return _convert_public_der_to_pem(der)
    elif secret_type == s.SecretType.CERTIFICATE:
        return _convert_certificate_der_to_pem(der)
    else:
        reason = u._("Secret type can not be converted to PEM")
        raise s.SecretGeneralException(reason=reason)


def _convert_private_pem_to_der(pem):
    private_key = serialization.load_pem_private_key(
        pem,
        password=None,
        backend=default_backend()
    )
    der = private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    return der


def _convert_private_der_to_pem(der):
    private_key = serialization.load_der_private_key(
        der,
        password=None,
        backend=default_backend()
    )
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    return pem


def _convert_public_pem_to_der(pem):
    public_key = serialization.load_pem_public_key(
        pem,
        backend=default_backend()
    )
    der = public_key.public_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    return der


def _convert_public_der_to_pem(der):
    public_key = serialization.load_der_public_key(
        der,
        backend=default_backend()
    )
    pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    return pem


def _convert_certificate_pem_to_der(pem):
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, pem)
    der = crypto.dump_certificate(crypto.FILETYPE_ASN1, cert)
    return der


def _convert_certificate_der_to_pem(der):
    cert = crypto.load_certificate(crypto.FILETYPE_ASN1, der)
    pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
    return pem



# Copyright (c) 2014 Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from oslo_config import cfg
import six
from stevedore import named

from barbican.common import config
from barbican.common import exception
from barbican.common import utils
from barbican import i18n as u
from barbican.plugin.util import multiple_backends
from barbican.plugin.util import utils as plugin_utils


_SECRET_STORE = None

CONF = config.new_config()
DEFAULT_PLUGIN_NAMESPACE = 'barbican.secretstore.plugin'
DEFAULT_PLUGINS = ['store_crypto']

store_opt_group = cfg.OptGroup(name='secretstore',
                               title='Secret Store Plugin Options')
store_opts = [
    cfg.StrOpt('namespace',
               default=DEFAULT_PLUGIN_NAMESPACE,
               help=u._('Extension namespace to search for plugins.')
               ),
    cfg.MultiStrOpt('enabled_secretstore_plugins',
                    default=DEFAULT_PLUGINS,
                    help=u._('List of secret store plugins to load.')
                    ),
    cfg.BoolOpt('enable_multiple_secret_stores',
                default=False,
                help=u._('Flag to enable multiple secret store plugin'
                         ' backend support. Default is False')
                ),
    cfg.ListOpt('stores_lookup_suffix',
                help=u._('List of suffix to use for looking up plugins which '
                         'are supported with multiple backend support.')
                )
]
CONF.register_group(store_opt_group)
CONF.register_opts(store_opts, group=store_opt_group)
config.parse_args(CONF)

config.set_module_config("secretstore", CONF)


def list_opts():
    yield store_opt_group, store_opts


class SecretStorePluginNotFound(exception.BarbicanHTTPException):
    """Raised when no plugins are installed."""

    client_message = u._("No plugin was found that could support your request")
    status_code = 400

    def __init__(self, plugin_name=None):
        if plugin_name:
            message = u._('Secret store plugin "{name}"'
                          ' not found.').format(name=plugin_name)
        else:
            message = u._("Secret store plugin not found.")
        super(SecretStorePluginNotFound, self).__init__(message)


class SecretStoreSupportedPluginNotFound(exception.BarbicanHTTPException):
    """Raised when no secret store supported plugin is found."""

    client_message = u._("Secret store supported plugin not found.")
    status_code = 400

    def __init__(self, key_spec):
        message = u._("Could not find a secret store plugin for storing "
                      "secret with algorithm '{alg}' and bit-length "
                      "'{len}'.").format(alg=key_spec.alg,
                                         len=key_spec.bit_length)
        super(SecretStoreSupportedPluginNotFound, self).__init__(
            message)


class SecretGenerateSupportedPluginNotFound(exception.BarbicanHTTPException):
    """Raised when no secret generate supported plugin is found."""

    client_message = u._("Secret generate supported plugin not found.")
    status_code = 400

    def __init__(self, key_spec):
        message = u._("Could not find a secret store plugin for generating "
                      "secret with algorithm '{alg}' and bit-length "
                      "'{len}'.").format(alg=key_spec.alg,
                                         len=key_spec.bit_length)
        super(SecretGenerateSupportedPluginNotFound, self).__init__(
            message)


class SecretContentTypeNotSupportedException(exception.BarbicanHTTPException):
    """Raised when support for payload content type is not available."""

    status_code = 400

    def __init__(self, content_type):
        super(SecretContentTypeNotSupportedException, self).__init__(
            u._("A Content-Type of '{content_type}' for secrets is "
                "not supported").format(
                    content_type=content_type)
        )
        self.content_type = content_type
        self.client_message = u._(
            "content-type of '{content_type}' not supported").format(
                content_type=content_type)


class SecretContentEncodingNotSupportedException(
        exception.BarbicanHTTPException):
    """Raised when support for payload content encoding is not available."""
    status_code = 400

    def __init__(self, content_encoding):
        super(SecretContentEncodingNotSupportedException, self).__init__(
            u._("Secret Content-Encoding of '{content_encoding}' "
                "not supported").format(
                    content_encoding=content_encoding)
        )
        self.content_encoding = content_encoding
        self.client_message = u._(
            "content-encoding of '{content_encoding}' not supported").format(
                content_encoding=content_encoding)


class SecretNoPayloadProvidedException(exception.BarbicanException):
    """Raised when secret information is not provided."""
    def __init__(self):
        super(SecretNoPayloadProvidedException, self).__init__(
            u._('No secret information provided to encrypt.')
        )


class SecretContentEncodingMustBeBase64(exception.BarbicanHTTPException):
    """Raised when encoding must be base64."""

    client_message = u._("Text-based binary secret payloads must "
                         "specify a content-encoding of 'base64'")
    status_code = 400

    def __init__(self):
        super(SecretContentEncodingMustBeBase64, self).__init__(
            u._("Encoding type must be 'base64' for text-based payloads.")
        )


class SecretGeneralException(exception.BarbicanException):
    """Raised when a system fault has occurred."""
    def __init__(self, reason=u._('Unknown')):
        super(SecretGeneralException, self).__init__(
            u._('Problem seen during crypto processing - '
                'Reason: {reason}').format(reason=reason)
        )
        self.reason = reason


class SecretPayloadDecodingError(exception.BarbicanHTTPException):
    """Raised when payload could not be decoded."""

    client_message = u._("Problem decoding payload")
    status_code = 400

    def __init__(self):
        super(SecretPayloadDecodingError, self).__init__(
            u._("Problem decoding payload")
        )


class SecretAcceptNotSupportedException(exception.BarbicanHTTPException):
    """Raised when requested decrypted content-type is not available."""

    client_message = u._("Wrong payload content-type")
    status_code = 406

    def __init__(self, accept):
        super(SecretAcceptNotSupportedException, self).__init__(
            u._("Secret Accept of '{accept}' not supported").format(
                accept=accept)
        )
        self.accept = accept


class SecretNotFoundException(exception.BarbicanHTTPException):
    """Raised when secret information could not be located."""

    client_message = u._("Secret not found.")
    status_code = 404

    def __init__(self):
        super(SecretNotFoundException, self).__init__(
            u._('No secret information found'))


class SecretAlgorithmNotSupportedException(exception.BarbicanHTTPException):
    """Raised when support for an algorithm is not available."""

    client_message = u._("Requested algorithm is not supported")
    status_code = 400

    def __init__(self, algorithm):
        super(SecretAlgorithmNotSupportedException, self).__init__(
            u._("Secret algorithm of '{algorithm}' not supported").format(
                algorithm=algorithm)
        )
        self.algorithm = algorithm


class GeneratePassphraseNotSupportedException(exception.BarbicanHTTPException):
    """Raised when generating keys encrypted by passphrase is not supported."""

    client_message = (
        u._("Generating keys encrypted with passphrases is not supported")
    )
    status_code = 400

    def __init__(self):
        super(GeneratePassphraseNotSupportedException, self).__init__(
            self.client_message
        )


class SecretStorePluginsNotConfigured(exception.BarbicanException):
    """Raised when there are no secret store plugins configured."""
    def __init__(self):
        super(SecretStorePluginsNotConfigured, self).__init__(
            u._('No secret store plugins have been configured')
        )


class StorePluginNotAvailableOrMisconfigured(exception.BarbicanException):
    """Raised when a plugin that was previously used can not be found."""
    def __init__(self, plugin_name):
        super(StorePluginNotAvailableOrMisconfigured, self).__init__(
            u._("The requested Store Plugin {plugin_name} is not "
                "currently available. This is probably a server "
                "misconfiguration.").format(
                plugin_name=plugin_name)
        )
        self.plugin_name = plugin_name


class SecretType(object):

    """Constant to define the symmetric key type.

    Used by getSecret to retrieve a symmetric key.
    """
    SYMMETRIC = "symmetric"
    """Constant to define the public key type. Used by getSecret to retrieve a
    public key.
    """
    PUBLIC = "public"
    """Constant to define the private key type. Used by getSecret to retrieve a
    private key.
    """
    PRIVATE = "private"
    """Constant to define the passphrase type. Used by getSecret to retrieve a
    passphrase."""
    PASSPHRASE = "passphrase"  # nosec
    """Constant to define the certificate type. Used by getSecret to retrieve a
    certificate."""
    CERTIFICATE = "certificate"
    """Constant to define the opaque date type. Used by getSecret to retrieve
    opaque data. Opaque data can be any kind of data. This data type signals to
    Barbican to just store the information and do not worry about the format or
    encoding. This is the default type if no type is specified by the user."""
    OPAQUE = utils.SECRET_TYPE_OPAQUE


class KeyAlgorithm(object):

    """Constant for the Diffie Hellman algorithm."""
    DIFFIE_HELLMAN = "diffie_hellman"
    """Constant for the DSA algorithm."""
    DSA = "dsa"
    """Constant for the RSA algorithm."""
    RSA = "rsa"
    """Constant for the Elliptic Curve algorithm."""
    EC = "ec"
    """Constant for the HMACSHA1 algorithm."""
    HMACSHA1 = "hmacsha1"
    """Constant for the HMACSHA256 algorithm."""
    HMACSHA256 = "hmacsha256"
    """Constant for the HMACSHA384 algorithm."""
    HMACSHA384 = "hmacsha384"
    """Constant for the HMACSHA512 algorithm."""
    HMACSHA512 = "hmacsha512"
    """List of asymmetric algorithms"""
    ASYMMETRIC_ALGORITHMS = [DIFFIE_HELLMAN, DSA, RSA, EC]
    """Constant for the AES algorithm."""
    AES = "aes"
    """Constant for the DES algorithm."""
    DES = "des"
    """Constant for the DESede (triple-DES) algorithm."""
    DESEDE = "desede"
    """List of symmetric algorithms"""
    SYMMETRIC_ALGORITHMS = [AES, DES, DESEDE, HMACSHA1,
                            HMACSHA256, HMACSHA384, HMACSHA512]


class KeySpec(object):
    """This object specifies the algorithm and bit length for a key."""

    def __init__(self, alg=None, bit_length=None, mode=None, passphrase=None):
        """Creates a new KeySpec.

        :param alg:algorithm for the key
        :param bit_length:bit length of the key
        :param mode:algorithm mode for the key
        :param passphrase:passphrase for the private_key
        """
        self.alg = alg
        self.bit_length = bit_length
        self.mode = mode  # TODO(john-wood-w) Paul, is 'mode' required?
        self.passphrase = passphrase


class SecretDTO(object):
    """This object is a secret data transfer object (DTO).

    This object encapsulates a key and attributes about the key. The attributes
    include a KeySpec that contains the algorithm and bit length. The
    attributes also include information on the encoding of the key.
    """

    # TODO(john-wood-w) Remove 'content_type' once secret normalization work is
    #  completed.
    def __init__(self, type, secret, key_spec, content_type,
                 transport_key=None):
        """Creates a new SecretDTO.

        The secret is stored in the secret parameter. In the future this
        DTO may include compression and key wrapping information.

        :param type: SecretType for secret
        :param secret: secret, as a base64-encoded string
        :param key_spec: KeySpec key specifications
        :param content_type: Content type of the secret, one of MIME
               types such as 'text/plain' or 'application/octet-stream'
        :param transport_key: presence of this parameter indicates that the
               secret has been encrypted using a transport key.  The transport
               key is a base64 encoded x509 transport certificate.
        """
        self.type = type or SecretType.OPAQUE
        self.secret = secret
        self.key_spec = key_spec
        self.content_type = content_type
        self.transport_key = transport_key


class AsymmetricKeyMetadataDTO(object):
    """This DTO encapsulates metadata(s) for asymmetric key components.

    These components are private_key_meta, public_key_meta and passphrase_meta.
    """

    def __init__(self, private_key_meta=None,
                 public_key_meta=None,
                 passphrase_meta=None):
        """Constructor for AsymmetricKeyMetadataDTO

        :param private_key_meta: private key metadata
        :param public_key_meta: public key metadata
        :param passphrase_meta: passphrase key metadata
        """
        self.private_key_meta = private_key_meta
        self.public_key_meta = public_key_meta
        self.passphrase_meta = passphrase_meta


@six.add_metaclass(abc.ABCMeta)
class SecretStoreBase(object):

    @abc.abstractmethod
    def get_plugin_name(self):
        """Gets user friendly plugin name.

        This plugin name is expected to be read from config file.
        There will be a default defined for plugin name which can be customized
        in specific deployment if needed.

        This name needs to be unique across a deployment.
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def generate_symmetric_key(self, key_spec):
        """Generate a new symmetric key and store it.

        Generates a new symmetric key and stores it in the secret store.
        A dictionary is returned that contains metadata about the newly created
        symmetric key. The dictionary of metadata is stored by Barbican and
        passed into other methods to aid the plugins. This can be useful for
        plugins that generate a unique ID in the external data store and use it
        to retrieve the key in the future. The returned dictionary may be empty
        if the SecretStore does not require it.

        :param key_spec: KeySpec that contains details on the type of key to
            generate
        :returns: an optional dictionary containing metadata about the key
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def generate_asymmetric_key(self, key_spec):
        """Generate a new asymmetric key pair and store it.

        Generates a new asymmetric key pair and stores it in the secret
        store. An object of type AsymmetricKeyMetadataDTO will be returned
        containing attributes of metadata for newly created key pairs.
        The metadata is stored by Barbican and passed into other methods
        to aid the plugins. This can be useful for plugins that generate
        a unique ID in the external data store and use it to retrieve the
        key pairs in the future.

        :param key_spec: KeySpec that contains details on the type of key to
            generate
        :returns: An object of type AsymmetricKeyMetadataDTO containing
            metadata about the key pair.
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def store_secret(self, secret_dto):
        """Stores a key.

        The SecretDTO contains the bytes of the secret and properties of the
        secret.  The SecretStore retrieves the secret bytes, stores them, and
        returns a dictionary of metadata about the secret.  This can be
        useful for plugins that generate a unique ID in the external data
        store and use it to retrieve the secret in the future. The returned
        dictionary may be empty if the SecretStore does not require it.

        :param secret_dto: SecretDTO for secret
        :returns: an optional dictionary containing metadata about the secret
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def get_secret(self, secret_type, secret_metadata):
        """Retrieves a secret from the secret store.

        Retrieves a secret from the secret store and returns a SecretDTO that
        contains the secret.

        The secret_metadata parameter is the metadata returned from one of the
        generate or store methods. This data is used by the plugins to retrieve
        the key.

        The secret_type parameter may be useful for secret stores to know the
        expected format of the secret. For instance if the type is
        SecretDTO.PRIVATE then a PKCS8 structure is returned. This way secret
        stores do not need to manage the secret type on their own.

        :param secret_type: secret type
        :param secret_metadata: secret metadata
        :returns: SecretDTO that contains secret
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def generate_supports(self, key_spec):
        """Returns a boolean indicating if the secret type is supported.

        This checks if the algorithm and bit length are supported by the
        generate methods. This is useful to call before calling
        generate_symmetric_key or generate_asymetric_key to see if the key type
        is supported before trying to generate it.

        :param key_spec: KeySpec that contains details on the algorithm and bit
            length
        :returns: boolean indicating if the algorithm is supported
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def delete_secret(self, secret_metadata):
        """Deletes a secret from the secret store.

        Deletes a secret from a secret store. It can no longer be referenced
        after this call.

        :param secret_metadata: secret_metadata
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def store_secret_supports(self, key_spec):
        """Returns a boolean indicating if the secret can be stored.

        Checks if the secret store can store the secret, give the attributes
        of the secret in the KeySpec. For example, some plugins may need to
        know the attributes in order to store the secret, but other plugins
        may be able to store the secret as a blob if no attributes are given.

        :param key_spec: KeySpec for the secret
        :returns: a boolean indicating if the secret can be stored
        """
        raise NotImplementedError  # pragma: no cover

    def get_transport_key(self):
        """Gets a transport key.

        Returns the current valid transport key associated with this plugin.
        The transport key is expected to be a base64 encoded x509 certificate
        containing a public key.  Admins are responsible for deleting old keys
        from the database using the DELETE method on the TransportKey resource.

        By default, returns None.  Plugins that support transport key
        wrapping should override this method.
        """
        return None

    def is_transport_key_current(self, transport_key):
        """Determines if the provided transport key is the current valid key

        Returns true if the transport key is the current valid transport key.
        If the key is not valid, then barbican core will request a new
        transport key from the plugin.

        Returns False by default.  Plugins that support transport key wrapping
        should override this method.
        """
        return False


def _enforce_extensions_configured(plugin_related_function):
    def _check_plugins_configured(self, *args, **kwargs):
        if not self.extensions:
            raise SecretStorePluginsNotConfigured()
        return plugin_related_function(self, *args, **kwargs)
    return _check_plugins_configured


class SecretStorePluginManager(named.NamedExtensionManager):
    def __init__(self, conf=CONF, invoke_args=(), invoke_kwargs={}):
        ss_conf = config.get_module_config('secretstore')
        plugin_names = self._get_internal_plugin_names(ss_conf)

        super(SecretStorePluginManager, self).__init__(
            ss_conf.secretstore.namespace,
            plugin_names,
            invoke_on_load=False,  # Defer creating plugins to utility below.
            invoke_args=invoke_args,
            invoke_kwds=invoke_kwargs,
            name_order=True  # extensions sorted as per order of plugin names
        )

        plugin_utils.instantiate_plugins(self, invoke_args, invoke_kwargs)

        multiple_backends.sync_secret_stores(self)

    @_enforce_extensions_configured
    def get_plugin_store(self, key_spec, plugin_name=None,
                         transport_key_needed=False, project_id=None):
        """Gets a secret store plugin.

        :param: plugin_name: set to plugin_name to get specific plugin
        :param: key_spec: KeySpec of key that will be stored
        :param: transport_key_needed: set to True if a transport
        key is required.
        :returns: SecretStoreBase plugin implementation
        """
        active_plugins = multiple_backends.get_applicable_store_plugins(
            self, project_id=project_id, existing_plugin_name=plugin_name)

        if plugin_name is not None:
            for plugin in active_plugins:
                if utils.generate_fullname_for(plugin) == plugin_name:
                    return plugin
            raise SecretStorePluginNotFound(plugin_name)

        if not transport_key_needed:
            for plugin in active_plugins:
                if plugin.store_secret_supports(key_spec):
                    return plugin

        else:
            for plugin in active_plugins:
                if (plugin.get_transport_key() is not None and
                        plugin.store_secret_supports(key_spec)):
                    return plugin

        raise SecretStoreSupportedPluginNotFound(key_spec)

    @_enforce_extensions_configured
    def get_plugin_retrieve_delete(self, plugin_name):
        """Gets a secret retrieve/delete plugin.

        If this function is being called, it is because we are trying to
        retrieve or delete an already stored secret. Thus, the plugin name is
        actually gotten from the plugin metadata that has already been stored
        in the database. So, in this case, if this plugin is not available,
        this might be due to a server misconfiguration.

        :returns: SecretStoreBase plugin implementation
        :raises: StorePluginNotAvailableOrMisconfigured: If the plugin wasn't
                 found it's because the plugin parameters were not properly
                 configured on the database side.
        """

        for plugin in plugin_utils.get_active_plugins(self):
            if utils.generate_fullname_for(plugin) == plugin_name:
                return plugin
        raise StorePluginNotAvailableOrMisconfigured(plugin_name)

    @_enforce_extensions_configured
    def get_plugin_generate(self, key_spec, project_id=None):
        """Gets a secret generate plugin.

        :param key_spec: KeySpec that contains details on the type of key to
        generate
        :returns: SecretStoreBase plugin implementation
        """

        active_plugins = multiple_backends.get_applicable_store_plugins(
            self, project_id=project_id, existing_plugin_name=None)

        for plugin in active_plugins:
            if plugin.generate_supports(key_spec):
                return plugin
        raise SecretGenerateSupportedPluginNotFound(key_spec)

    def _get_internal_plugin_names(self, secretstore_conf):
        """Gets plugin names used for loading via stevedore.

        When multiple secret store support is enabled, then secret store plugin
        names are read via updated configuration structure. If not enabled,
        then it reads MultiStr property in 'secretstore' config section.
        """
        # to cache default global secret store value on first use
        self.global_default_store_dict = None
        if utils.is_multiple_backends_enabled():
            self.parsed_stores = multiple_backends.\
                read_multiple_backends_config()
            plugin_names = [store.store_plugin for store in self.parsed_stores
                            if store.store_plugin]
        else:
            plugin_names = secretstore_conf.secretstore.\
                enabled_secretstore_plugins
        return plugin_names


def get_manager():
    global _SECRET_STORE
    if not _SECRET_STORE:
        _SECRET_STORE = SecretStorePluginManager()
    return _SECRET_STORE



"""
Copyright 2014-2015 Rackspace

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from barbican.tests import utils
from functionaltests.api.v1.behaviors import base_behaviors
from functionaltests.api.v1.models import container_models


class ContainerBehaviors(base_behaviors.BaseBehaviors):

    def create_container(self, model, extra_headers=None,
                         user_name=None, admin=None):
        """Create a container from the data in the model.

        :param model: The metadata used to create the container
        :param extra_headers: Headers used to create the container
        :param user_name: The user name used to create the container
        :param admin: The user with permissions to delete the container
        :return: A tuple containing the response from the create
        and the href to the newly created container
        """

        resp = self.client.post('containers', request_model=model,
                                extra_headers=extra_headers,
                                user_name=user_name)

        returned_data = self.get_json(resp)
        container_ref = returned_data.get('container_ref')
        if container_ref:
            if admin is None:
                admin = user_name
            self.created_entities.append((container_ref, admin))
        return resp, container_ref

    def get_container(self, container_ref, extra_headers=None, user_name=None):
        """Handles getting a single container

        :param container_ref: Reference to the container to be retrieved
        :param extra_headers: Headers used to get the container
        :param user_name: The user name used to get the container

        :return: The response of the GET.
        """
        resp = self.client.get(
            container_ref, response_model_type=container_models.ContainerModel,
            user_name=user_name, extra_headers=extra_headers)

        return resp

    def get_containers(self, limit=10, offset=0, filter=None,
                       extra_headers=None, user_name=None):
        """Handles getting a list of containers.

        :param limit: limits number of returned containers
        :param offset: represents how many records to skip before retrieving
            the list
        :param filter: allows you to filter results based on name
        :param extra_headers: Extra headers used to retrieve a list of
            containers
        :param user_name: The user name used to get the list

        :return: Returns the response, a list of container models, and
            references to the next and previous list of containers.
        """
        params = {'limit': limit, 'offset': offset}

        if filter:
            params['name'] = filter

        resp = self.client.get('containers', params=params,
                               extra_headers=extra_headers,
                               user_name=user_name)

        container_list = self.get_json(resp)

        containers, next_ref, prev_ref = self.client.get_list_of_models(
            container_list, container_models.ContainerModel)

        return resp, containers, next_ref, prev_ref

    def delete_container(self, container_ref, extra_headers=None,
                         expected_fail=False, user_name=None):
        """Handles deleting a containers.

        :param container_ref: Reference of the container to be deleted
        :param extra_headers: Any additional headers needed.
        :param expected_fail: If there is a negative test, this should be
            marked true if you are trying to delete a container that does
            not exist.
        :param user_name: The user name used to delete the container
        :return: Response of the delete.
        """
        resp = self.client.delete(container_ref, extra_headers,
                                  user_name=user_name)

        if not expected_fail:
            for item in self.created_entities:
                if item[0] == container_ref:
                    self.created_entities.remove(item)

        return resp

    def delete_all_created_containers(self):
        """Delete all of the containers that we have created."""
        entities = list(self.created_entities)
        for (container_ref, admin) in entities:
            self.delete_container(container_ref, user_name=admin)

    def update_container(self, container_ref, user_name=None):
        """Attempt to update a container (which is an invalid operation)

        Update (HTTP PUT) is not supported against a container resource, so
        issuing this call should fail.

        :param container_ref: Reference of the container to be updated
        :param user_name: The user name used to update the container
        :return: Response of the update.
        """

        resp = self.client.put(container_ref, user_name=user_name)

        return resp

    def delete_all_containers_for_user(self, user_name):
        '''Delete all of the containers for the specified user'''
        response, containers, next_ref, prev_ref = self.get_containers(
            user_name=user_name)
        container_refs_to_delete = []
        while len(containers) > 0:
            for container in containers:
                container_refs_to_delete.append(container.container_ref)
            if next_ref:
                limit, offset = utils.get_limit_and_offset_from_ref(next_ref)
                response, containers, next_ref, prev = self.get_containers(
                    limit=limit, offset=offset, user_name=user_name)
            else:
                break

        for container_ref in container_refs_to_delete:
            self.delete_container(container_ref=container_ref,
                                  user_name=user_name)



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Ocata rebase

Revision ID: 39cf2e645cba
Revises: None
Create Date: 2020-09-23 07:55:03.238363

"""

# revision identifiers, used by Alembic.
revision = '39cf2e645cba'
down_revision = None

from alembic import op
import sqlalchemy as sa

import barbican


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'projects',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('external_id', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('external_id'),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'secret_stores',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('store_plugin', sa.String(length=255), nullable=False),
        sa.Column('crypto_plugin', sa.String(length=255), nullable=True),
        sa.Column('global_default', sa.Boolean(), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', name='_secret_stores_name_uc'),
        sa.UniqueConstraint('store_plugin', 'crypto_plugin',
                            name='_secret_stores_plugin_names_uc')
    )
    op.create_table(
        'transport_keys',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('plugin_name', sa.String(length=255), nullable=False),
        sa.Column('transport_key', sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'certificate_authorities',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('plugin_name', sa.String(length=255), nullable=False),
        sa.Column('plugin_ca_id', sa.Text(), nullable=False),
        sa.Column('expiration', sa.DateTime(), nullable=True),
        sa.Column('creator_id', sa.String(length=255), nullable=True),
        sa.Column('project_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'],
                                name='cas_project_fk'),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'containers',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('type', sa.Enum('generic', 'rsa', 'dsa', 'certificate',
                                  name='container_types'), nullable=True),
        sa.Column('project_id', sa.String(length=36), nullable=False),
        sa.Column('creator_id', sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'],
                                name='containers_project_fk'),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_containers_project_id'),
                    'containers', ['project_id'], unique=False)
    op.create_table(
        'kek_data',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('plugin_name', sa.String(length=255), nullable=False),
        sa.Column('kek_label', sa.String(length=255), nullable=True),
        sa.Column('project_id', sa.String(length=36), nullable=False),
        sa.Column('active', sa.Boolean(), nullable=False),
        sa.Column('bind_completed', sa.Boolean(), nullable=False),
        sa.Column('algorithm', sa.String(length=255), nullable=True),
        sa.Column('bit_length', sa.Integer(), nullable=True),
        sa.Column('mode', sa.String(length=255), nullable=True),
        sa.Column('plugin_meta', sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'],
                                name='kek_data_project_fk'),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_kek_data_project_id'),
                    'kek_data', ['project_id'], unique=False)
    op.create_table(
        'project_quotas',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('project_id', sa.String(length=36), nullable=False),
        sa.Column('secrets', sa.Integer(), nullable=True),
        sa.Column('orders', sa.Integer(), nullable=True),
        sa.Column('containers', sa.Integer(), nullable=True),
        sa.Column('consumers', sa.Integer(), nullable=True),
        sa.Column('cas', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'],
                                name='project_quotas_fk'),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_project_quotas_project_id'), 'project_quotas',
                    ['project_id'], unique=False)
    op.create_table(
        'project_secret_store',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('secret_store_id', sa.String(length=36), nullable=False),
        sa.Column('project_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
        sa.ForeignKeyConstraint(['secret_store_id'], ['secret_stores.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('project_id',
                            name='_project_secret_store_project_uc')
    )
    op.create_index(op.f('ix_project_secret_store_project_id'),
                    'project_secret_store', ['project_id'], unique=False)
    op.create_index(op.f('ix_project_secret_store_secret_store_id'),
                    'project_secret_store', ['secret_store_id'],
                    unique=False)
    op.create_table(
        'secrets',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('secret_type', sa.String(length=255),
                  server_default='opaque', nullable=True),
        sa.Column('expiration', sa.DateTime(), nullable=True),
        sa.Column('algorithm', sa.String(length=255), nullable=True),
        sa.Column('bit_length', sa.Integer(), nullable=True),
        sa.Column('mode', sa.String(length=255), nullable=True),
        sa.Column('creator_id', sa.String(length=255), nullable=True),
        sa.Column('project_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'],
                                name='secrets_project_fk'),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_secrets_project_id'), 'secrets', ['project_id'],
                    unique=False)
    op.create_table(
        'certificate_authority_metadata',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('key', sa.String(length=255), nullable=False),
        sa.Column('value', sa.Text(), nullable=False),
        sa.Column('ca_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['ca_id'], ['certificate_authorities.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('ca_id', 'key',
                            name='_certificate_authority_metadatum_uc')
    )
    op.create_index(op.f('ix_certificate_authority_metadata_ca_id'),
                    'certificate_authority_metadata', ['ca_id'],
                    unique=False)
    op.create_index(op.f('ix_certificate_authority_metadata_key'),
                    'certificate_authority_metadata', ['key'], unique=False)
    op.create_table(
        'container_acls',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('container_id', sa.String(length=36), nullable=False),
        sa.Column('operation', sa.String(length=255), nullable=False),
        sa.Column('project_access', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(['container_id'], ['containers.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('container_id', 'operation',
                            name='_container_acl_operation_uc')
    )
    op.create_index(op.f('ix_container_acls_container_id'), 'container_acls',
                    ['container_id'], unique=False)
    op.create_table(
        'container_consumer_metadata',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('container_id', sa.String(length=36), nullable=False),
        sa.Column('project_id', sa.String(length=36), nullable=True),
        sa.Column('name', sa.String(length=36), nullable=True),
        sa.Column('URL', sa.String(length=255), nullable=True),
        sa.Column('data_hash', sa.CHAR(length=64), nullable=True),
        sa.ForeignKeyConstraint(['container_id'], ['containers.id'], ),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('data_hash',
                            name='_consumer_hashed_container_name_url_uc')
    )
    op.create_index(op.f('ix_container_consumer_metadata_container_id'),
                    'container_consumer_metadata', ['container_id'],
                    unique=False)
    op.create_index(op.f('ix_container_consumer_metadata_project_id'),
                    'container_consumer_metadata', ['project_id'],
                    unique=False)
    op.create_index('values_index', 'container_consumer_metadata',
                    ['container_id', 'name', 'URL'], unique=False)
    op.create_table(
        'container_secret',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('container_id', sa.String(length=36), nullable=False),
        sa.Column('secret_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['container_id'], ['containers.id'], ),
        sa.ForeignKeyConstraint(['secret_id'], ['secrets.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('container_id', 'secret_id', 'name',
                            name='_container_secret_name_uc')
    )
    op.create_index(op.f('ix_container_secret_container_id'),
                    'container_secret', ['container_id'], unique=False)
    op.create_index(op.f('ix_container_secret_secret_id'), 'container_secret',
                    ['secret_id'], unique=False)
    op.create_table(
        'encrypted_data',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('content_type', sa.String(length=255), nullable=True),
        sa.Column('secret_id', sa.String(length=36), nullable=False),
        sa.Column('kek_id', sa.String(length=36), nullable=False),
        sa.Column('cypher_text', sa.Text(), nullable=True),
        sa.Column('kek_meta_extended', sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(['kek_id'], ['kek_data.id'], ),
        sa.ForeignKeyConstraint(['secret_id'], ['secrets.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_encrypted_data_kek_id'), 'encrypted_data',
                    ['kek_id'], unique=False)
    op.create_index(op.f('ix_encrypted_data_secret_id'), 'encrypted_data',
                    ['secret_id'], unique=False)
    op.create_table(
        'orders',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('type', sa.String(length=255), nullable=False),
        sa.Column('project_id', sa.String(length=36), nullable=False),
        sa.Column('error_status_code', sa.String(length=16), nullable=True),
        sa.Column('error_reason', sa.String(length=255), nullable=True),
        sa.Column('meta', barbican.model.models.JsonBlob(), nullable=True),
        sa.Column('secret_id', sa.String(length=36), nullable=True),
        sa.Column('container_id', sa.String(length=36), nullable=True),
        sa.Column('sub_status', sa.String(length=36), nullable=True),
        sa.Column('sub_status_message', sa.String(length=255), nullable=True),
        sa.Column('creator_id', sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(['container_id'], ['containers.id'], ),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'],
                                name='orders_project_fk'),
        sa.ForeignKeyConstraint(['secret_id'], ['secrets.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_orders_container_id'), 'orders', ['container_id'],
                    unique=False)
    op.create_index(op.f('ix_orders_project_id'), 'orders', ['project_id'],
                    unique=False)
    op.create_index(op.f('ix_orders_secret_id'), 'orders', ['secret_id'],
                    unique=False)
    op.create_table(
        'preferred_certificate_authorities',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('project_id', sa.String(length=36), nullable=False),
        sa.Column('ca_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['ca_id'], ['certificate_authorities.id'],
                                name='preferred_certificate_authorities_fk'),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_preferred_certificate_authorities_ca_id'),
                    'preferred_certificate_authorities', ['ca_id'],
                    unique=False)
    op.create_index(op.f('ix_preferred_certificate_authorities_project_id'),
                    'preferred_certificate_authorities', ['project_id'],
                    unique=True)
    op.create_table(
        'project_certificate_authorities',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('project_id', sa.String(length=36), nullable=False),
        sa.Column('ca_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['ca_id'], ['certificate_authorities.id'], ),
        sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('project_id', 'ca_id',
                            name='_project_certificate_authority_uc'))
    op.create_index(op.f('ix_project_certificate_authorities_ca_id'),
                    'project_certificate_authorities', ['ca_id'],
                    unique=False)
    op.create_index(op.f('ix_project_certificate_authorities_project_id'),
                    'project_certificate_authorities', ['project_id'],
                    unique=False)
    op.create_table(
        'secret_acls',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('secret_id', sa.String(length=36), nullable=False),
        sa.Column('operation', sa.String(length=255), nullable=False),
        sa.Column('project_access', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(['secret_id'], ['secrets.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('secret_id', 'operation',
                            name='_secret_acl_operation_uc')
    )
    op.create_index(op.f('ix_secret_acls_secret_id'), 'secret_acls',
                    ['secret_id'], unique=False)
    op.create_table(
        'secret_store_metadata',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('key', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=False),
        sa.Column('secret_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['secret_id'], ['secrets.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_secret_store_metadata_secret_id'),
                    'secret_store_metadata', ['secret_id'], unique=False)
    op.create_table(
        'secret_user_metadata',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('key', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=False),
        sa.Column('secret_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['secret_id'], ['secrets.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('secret_id', 'key', name='_secret_key_uc')
    )
    op.create_index(op.f('ix_secret_user_metadata_secret_id'),
                    'secret_user_metadata', ['secret_id'], unique=False)
    op.create_table(
        'container_acl_users',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('acl_id', sa.String(length=36), nullable=False),
        sa.Column('user_id', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['acl_id'], ['container_acls.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('acl_id', 'user_id', name='_container_acl_user_uc')
    )
    op.create_index(op.f('ix_container_acl_users_acl_id'),
                    'container_acl_users', ['acl_id'], unique=False)
    op.create_table(
        'order_barbican_metadata',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('order_id', sa.String(length=36), nullable=False),
        sa.Column('key', sa.String(length=255), nullable=False),
        sa.Column('value', sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(['order_id'], ['orders.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_order_barbican_metadata_order_id'),
                    'order_barbican_metadata', ['order_id'], unique=False)
    op.create_table(
        'order_plugin_metadata',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('order_id', sa.String(length=36), nullable=False),
        sa.Column('key', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['order_id'], ['orders.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_order_plugin_metadata_order_id'),
                    'order_plugin_metadata', ['order_id'], unique=False)
    op.create_table(
        'order_retry_tasks',
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('order_id', sa.String(length=36), nullable=False),
        sa.Column('retry_task', sa.Text(), nullable=False),
        sa.Column('retry_at', sa.DateTime(), nullable=False),
        sa.Column('retry_args', barbican.model.models.JsonBlob(),
                  nullable=False),
        sa.Column('retry_kwargs', barbican.model.models.JsonBlob(),
                  nullable=False),
        sa.Column('retry_count', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['order_id'], ['orders.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_order_retry_tasks_order_id'), 'order_retry_tasks',
                    ['order_id'], unique=False)
    op.create_table(
        'secret_acl_users',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.Column('status', sa.String(length=20), nullable=False),
        sa.Column('acl_id', sa.String(length=36), nullable=False),
        sa.Column('user_id', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['acl_id'], ['secret_acls.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('acl_id', 'user_id', name='_secret_acl_user_uc')
    )
    op.create_index(op.f('ix_secret_acl_users_acl_id'), 'secret_acl_users',
                    ['acl_id'], unique=False)
    # ### end Alembic commands ###



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from barbican.common import exception
from barbican.model import repositories
from barbican.tests import database_utils


class WhenTestingTransportKeyRepository(database_utils.RepositoryTestCase):

    def setUp(self):
        super(WhenTestingTransportKeyRepository, self).setUp()
        self.repo = repositories.TransportKeyRepo()

    def test_should_raise_no_result_found_with_plugin_name(self):
        session = self.repo.get_session()

        self.assertRaises(
            exception.NotFound,
            self.repo.get_by_create_date,
            plugin_name="plugin",
            session=session,
            suppress_exception=False)

    def test_should_raise_no_result_found_no_plugin_name(self):
        session = self.repo.get_session()

        self.assertRaises(
            exception.NotFound,
            self.repo.get_by_create_date,
            session=session,
            suppress_exception=False)



# Copyright (c) 2014 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import copy
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
import datetime
import os
from oslo_utils import uuidutils
import time

import pki

subcas_available = True
try:
    import pki.authority as authority
    import pki.feature as feature
except ImportError:
    subcas_available = False

import pki.cert
import pki.client
import pki.crypto as cryptoutil
import pki.key as key
import pki.kra
import pki.profile
from requests import exceptions as request_exceptions
import six

from barbican.common import exception
from barbican.common import utils
from barbican import i18n as u

# we want to keep the dogtag config options separated. That way we
# do not need to import every dogtag requirement to generate the
# sample config
import barbican.plugin.dogtag_config_opts  # noqa
import barbican.plugin.interface.certificate_manager as cm
import barbican.plugin.interface.secret_store as sstore

# reuse the conf object to not call config.new_config() twice
CONF = barbican.plugin.dogtag_config_opts.CONF
LOG = utils.getLogger(__name__)

CERT_HEADER = "-----BEGIN CERTIFICATE-----"
CERT_FOOTER = "-----END CERTIFICATE-----"

KRA_TRANSPORT_NICK = "KRA transport cert"


def _create_nss_db_if_needed(nss_db_path, nss_password):
    """Creates NSS DB if it's not setup already

    :returns: True or False whether the database was created or not.
    """
    if not os.path.exists(nss_db_path):
        cryptoutil.NSSCryptoProvider.setup_database(
            nss_db_path, nss_password, over_write=True)
        return True
    else:
        LOG.info("The nss_db_path provided already exists, so the "
                 "database is assumed to be already set up.")
        return False


def _setup_nss_db_services(conf):
    """Sets up NSS Crypto functions

    This sets up the NSSCryptoProvider and the database it needs for it to
    store certificates. If the path specified in the configuration is already
    existent, it will assume that the database is already setup.

    This will also import the transport cert needed by the KRA if the NSS DB
    was created.
    """
    nss_db_path, nss_password = (conf.dogtag_plugin.nss_db_path,
                                 conf.dogtag_plugin.nss_password)
    if nss_db_path is None:
        LOG.warning("nss_db_path was not provided so the crypto "
                    "provider functions were not initialized.")
        return None
    if nss_password is None:
        raise ValueError(u._("nss_password is required"))
    if type(nss_password) is not six.binary_type:
        # Password needs to be a bytes object in Python 3
        nss_password = nss_password.encode('UTF-8')

    nss_db_created = _create_nss_db_if_needed(nss_db_path, nss_password)
    crypto = cryptoutil.NSSCryptoProvider(nss_db_path, nss_password)
    if nss_db_created:
        _import_kra_transport_cert_to_nss_db(conf, crypto)

    return crypto


def _import_kra_transport_cert_to_nss_db(conf, crypto):
    try:
        connection = create_connection(conf, 'kra')
        kraclient = pki.kra.KRAClient(connection, crypto)
        systemcert_client = kraclient.system_certs

        transport_cert = systemcert_client.get_transport_cert()
        crypto.import_cert(KRA_TRANSPORT_NICK, transport_cert, ",,")
    except Exception as e:
        LOG.debug("Error importing KRA transport cert.", exc_info=True)
        LOG.error("Error in importing transport cert."
                  " KRA may not be enabled: %s", e)


def create_connection(conf, subsystem_path):
    pem_path = conf.dogtag_plugin.pem_path
    if pem_path is None:
        raise ValueError(u._("pem_path is required"))
    # port is string type in PKIConnection
    connection = pki.client.PKIConnection(
        'https',
        conf.dogtag_plugin.dogtag_host,
        str(conf.dogtag_plugin.dogtag_port),
        subsystem_path)
    connection.set_authentication_cert(pem_path)
    return connection

crypto = _setup_nss_db_services(CONF)
if crypto:
    crypto.initialize()


class DogtagPluginAlgorithmException(exception.BarbicanException):
    message = u._("Invalid algorithm passed in")


class DogtagPluginNotSupportedException(exception.NotSupported):
    message = u._("Operation not supported by Dogtag Plugin")

    def __init__(self, msg=None):
        if not msg:
            message = self.message
        else:
            message = msg

        super(DogtagPluginNotSupportedException, self).__init__(message)


class DogtagPluginArchivalException(exception.BarbicanException):
    message = u._("Key archival failed.  Error returned from KRA.")


class DogtagPluginGenerationException(exception.BarbicanException):
    message = u._("Key generation failed.  Error returned from KRA.")


class DogtagKRAPlugin(sstore.SecretStoreBase):
    """Implementation of the secret store plugin with KRA as the backend."""

    # metadata constants
    ALG = "alg"
    BIT_LENGTH = "bit_length"
    GENERATED = "generated"
    KEY_ID = "key_id"
    SECRET_MODE = "secret_mode"  # nosec
    PASSPHRASE_KEY_ID = "passphrase_key_id"  # nosec
    CONVERT_TO_PEM = "convert_to_pem"

    # string constants
    DSA_PRIVATE_KEY_HEADER = '-----BEGIN DSA PRIVATE KEY-----'
    DSA_PRIVATE_KEY_FOOTER = '-----END DSA PRIVATE KEY-----'
    DSA_PUBLIC_KEY_HEADER = '-----BEGIN DSA PUBLIC KEY-----'
    DSA_PUBLIC_KEY_FOOTER = '-----END DSA PUBLIC KEY-----'

    def __init__(self, conf=CONF):
        """Constructor - create the keyclient."""
        LOG.debug("starting DogtagKRAPlugin init")
        connection = create_connection(conf, 'kra')

        # create kraclient
        kraclient = pki.kra.KRAClient(connection, crypto)
        self.keyclient = kraclient.keys

        self.keyclient.set_transport_cert(KRA_TRANSPORT_NICK)
        self.plugin_name = conf.dogtag_plugin.plugin_name
        self.retries = conf.dogtag_plugin.retries

        LOG.debug("completed DogtagKRAPlugin init")

    def get_plugin_name(self):
        return self.plugin_name

    def store_secret(self, secret_dto):
        """Store a secret in the KRA

        If secret_dto.transport_key is not None, then we expect
        secret_dto.secret to include a base64 encoded PKIArchiveOptions
        structure as defined in section 6.4 of RFC 2511. This package contains
        a transport key wrapped session key, the session key wrapped secret
        and parameters to specify the symmetric key wrapping.

        Otherwise, the data is unencrypted and we use a call to archive_key()
        to have the Dogtag KRA client generate the relevant session keys.

        The secret_dto contains additional information on the type of secret
        that is being stored.  We will use that shortly.  For, now, lets just
        assume that its all PASS_PHRASE_TYPE

        Returns a dict with the relevant metadata (which in this case is just
        the key_id
        """
        data_type = key.KeyClient.PASS_PHRASE_TYPE
        key_id = None

        attempts = 0
        offset_time = 1
        while attempts <= self.retries and key_id is None:
            client_key_id = uuidutils.generate_uuid(dashed=False)
            if secret_dto.transport_key is not None:
                # TODO(alee-3) send the transport key with the archival request
                # once the Dogtag Client API changes.
                response = self.keyclient.archive_pki_options(
                    client_key_id,
                    data_type,
                    secret_dto.secret,
                    key_algorithm=None,
                    key_size=None)
            else:
                response = self.keyclient.archive_key(
                    client_key_id,
                    data_type,
                    secret_dto.secret,
                    key_algorithm=None,
                    key_size=None)

            key_id = response.get_key_id()

            if key_id is None:
                LOG.warning("key_id is None.  attempts: {}".format(attempts))
                attempts += 1
                time.sleep(offset_time)
                offset_time += 1

        if key_id is None:
            raise DogtagPluginArchivalException

        meta_dict = {DogtagKRAPlugin.KEY_ID: key_id}

        self._store_secret_attributes(meta_dict, secret_dto)
        return meta_dict

    def get_secret(self, secret_type, secret_metadata):
        """Retrieve a secret from the KRA

        The secret_metadata is simply the dict returned by a store_secret() or
        get_secret() call.  We will extract the key_id from this dict.

        Note: There are two ways to retrieve secrets from the KRA.

        The first method calls retrieve_key without a wrapping key.  This
        relies on the KRA client to generate a wrapping key (and wrap it with
        the KRA transport cert), and is completely transparent to the
        Barbican server.  What is returned to the caller is the
        unencrypted secret.

        The second way is to provide a wrapping key that would be generated
        on the barbican client.  That way only the client will be
        able to unwrap the secret.  This wrapping key is provided in the
        secret_metadata by Barbican core.

        Format/Type of the secret returned in the SecretDTO object.
        -----------------------------------------------------------
        The type of the secret returned is always dependent on the way it is
        stored using the store_secret method.

        In case of strings - like passphrase/PEM strings, the return will be a
        string.

        In case of binary data - the return will be the actual binary data.

        In case of retrieving an asymmetric key that is generated using the
        dogtag plugin, then the binary representation of, the asymmetric key in
        PEM format, is returned
        """
        key_id = secret_metadata[DogtagKRAPlugin.KEY_ID]

        key_spec = sstore.KeySpec(
            alg=secret_metadata.get(DogtagKRAPlugin.ALG, None),
            bit_length=secret_metadata.get(DogtagKRAPlugin.BIT_LENGTH, None),
            mode=secret_metadata.get(DogtagKRAPlugin.SECRET_MODE, None),
            passphrase=None
        )

        generated = secret_metadata.get(DogtagKRAPlugin.GENERATED, False)

        passphrase = self._get_passphrase_for_a_private_key(
            secret_type, secret_metadata, key_spec)

        recovered_key = None
        twsk = DogtagKRAPlugin._get_trans_wrapped_session_key(secret_type,
                                                              secret_metadata)

        if DogtagKRAPlugin.CONVERT_TO_PEM in secret_metadata:
            # Case for returning the asymmetric keys generated in KRA.
            # Asymmetric keys generated in KRA are not generated in PEM format.
            # This marker DogtagKRAPlugin.CONVERT_TO_PEM is set in the
            # secret_metadata for asymmetric keys generated in KRA to
            # help convert the returned private/public keys to PEM format and
            # eventually return the binary data of the keys in PEM format.

            if secret_type == sstore.SecretType.PUBLIC:
                # Public key should be retrieved using the get_key_info method
                # as it is treated as an attribute of the asymmetric key pair
                # stored in the KRA database.

                key_info = self.keyclient.get_key_info(key_id)
                recovered_key = serialization.load_der_public_key(
                    key_info.public_key,
                    backend=default_backend()
                ).public_bytes(
                    serialization.Encoding.PEM,
                    serialization.PublicFormat.PKCS1)

            elif secret_type == sstore.SecretType.PRIVATE:
                key_data = self.keyclient.retrieve_key(key_id)
                private_key = serialization.load_der_private_key(
                    key_data.data,
                    password=None,
                    backend=default_backend()
                )

                if passphrase is not None:
                    e_alg = serialization.BestAvailableEncryption(passphrase)
                else:
                    e_alg = serialization.NoEncryption()

                recovered_key = private_key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=e_alg
                )
        else:
            # TODO(alee-3) send transport key as well when dogtag client API
            # changes in case the transport key has changed.
            key_data = self.keyclient.retrieve_key(key_id, twsk)
            if twsk:
                # The data returned is a byte array.
                recovered_key = key_data.encrypted_data
            else:
                recovered_key = key_data.data

        # TODO(alee) remove final field when content_type is removed
        # from secret_dto

        if generated:
            recovered_key = base64.b64encode(recovered_key)

        ret = sstore.SecretDTO(
            type=secret_type,
            secret=recovered_key,
            key_spec=key_spec,
            content_type=None,
            transport_key=None)

        return ret

    def delete_secret(self, secret_metadata):
        """Delete a secret from the KRA

        There is currently no way to delete a secret in Dogtag.
        We will be implementing such a method shortly.
        """
        pass

    def generate_symmetric_key(self, key_spec):
        """Generate a symmetric key

        This calls generate_symmetric_key() on the KRA passing in the
        algorithm, bit_length and id (used as the client_key_id) from
        the secret.  The remaining parameters are not used.

        Returns a metadata object that can be used for retrieving the secret.
        """

        usages = [key.SymKeyGenerationRequest.DECRYPT_USAGE,
                  key.SymKeyGenerationRequest.ENCRYPT_USAGE]

        algorithm = self._map_algorithm(key_spec.alg.lower())

        if algorithm is None:
            raise DogtagPluginAlgorithmException
        passphrase = key_spec.passphrase
        if passphrase:
            raise DogtagPluginNotSupportedException(
                u._("Passphrase encryption is not supported for symmetric"
                    " key generating algorithms."))

        key_id = None
        attempts = 0
        offset_time = 1
        while attempts <= self.retries and key_id is None:
            client_key_id = uuidutils.generate_uuid()
            response = self.keyclient.generate_symmetric_key(
                client_key_id,
                algorithm,
                key_spec.bit_length,
                usages)
            key_id = response.get_key_id()

            if key_id is None:
                LOG.warning("generate_symkey: key_id is None.  attempts: {}"
                            .format(attempts))
                attempts += 1
                time.sleep(offset_time)
                offset_time += 1

        if key_id is None:
            raise DogtagPluginGenerationException

        # Barbican expects stored keys to be base 64 encoded.  We need to
        # add flag to the keyclient.generate_symmetric_key() call above
        # to ensure that the key that is stored is base64 encoded.
        #
        # As a workaround until that update is available, we will store a
        # parameter "generated"  to indicate that the response must be base64
        # encoded on retrieval.  Note that this will not work for transport
        # key encoded data.
        return {DogtagKRAPlugin.ALG: key_spec.alg,
                DogtagKRAPlugin.BIT_LENGTH: key_spec.bit_length,
                DogtagKRAPlugin.KEY_ID: response.get_key_id(),
                DogtagKRAPlugin.GENERATED: True}

    def generate_asymmetric_key(self, key_spec):
        """Generate an asymmetric key.

        Note that barbican expects all secrets to be base64 encoded.
        """

        usages = [key.AsymKeyGenerationRequest.DECRYPT_USAGE,
                  key.AsymKeyGenerationRequest.ENCRYPT_USAGE]

        client_key_id = uuidutils.generate_uuid()
        algorithm = self._map_algorithm(key_spec.alg.lower())
        passphrase = key_spec.passphrase

        if algorithm is None:
            raise DogtagPluginAlgorithmException

        passphrase_key_id = None
        passphrase_metadata = None
        if passphrase:
            if algorithm == key.KeyClient.DSA_ALGORITHM:
                raise DogtagPluginNotSupportedException(
                    u._("Passphrase encryption is not "
                        "supported for DSA algorithm")
                )

            stored_passphrase_info = self.keyclient.archive_key(
                uuidutils.generate_uuid(),
                self.keyclient.PASS_PHRASE_TYPE,
                base64.b64encode(passphrase))

            passphrase_key_id = stored_passphrase_info.get_key_id()
            passphrase_metadata = {
                DogtagKRAPlugin.KEY_ID: passphrase_key_id
            }

        # Barbican expects stored keys to be base 64 encoded.  We need to
        # add flag to the keyclient.generate_asymmetric_key() call above
        # to ensure that the key that is stored is base64 encoded.
        #
        # As a workaround until that update is available, we will store a
        # parameter "generated"  to indicate that the response must be base64
        # encoded on retrieval.  Note that this will not work for transport
        # key encoded data.

        response = self.keyclient.generate_asymmetric_key(
            client_key_id,
            algorithm,
            key_spec.bit_length,
            usages)

        public_key_metadata = {
            DogtagKRAPlugin.ALG: key_spec.alg,
            DogtagKRAPlugin.BIT_LENGTH: key_spec.bit_length,
            DogtagKRAPlugin.KEY_ID: response.get_key_id(),
            DogtagKRAPlugin.CONVERT_TO_PEM: "true",
            DogtagKRAPlugin.GENERATED: True
        }

        private_key_metadata = {
            DogtagKRAPlugin.ALG: key_spec.alg,
            DogtagKRAPlugin.BIT_LENGTH: key_spec.bit_length,
            DogtagKRAPlugin.KEY_ID: response.get_key_id(),
            DogtagKRAPlugin.CONVERT_TO_PEM: "true",
            DogtagKRAPlugin.GENERATED: True
        }

        if passphrase_key_id:
            private_key_metadata[DogtagKRAPlugin.PASSPHRASE_KEY_ID] = (
                passphrase_key_id
            )

        return sstore.AsymmetricKeyMetadataDTO(private_key_metadata,
                                               public_key_metadata,
                                               passphrase_metadata)

    def generate_supports(self, key_spec):
        """Key generation supported?

        Specifies whether the plugin supports key generation with the
        given key_spec.

        For now, we will just check the algorithm.  When dogtag adds a
        call to check the bit length as well, we will use that call to
        take advantage of the bit_length information
        """
        return self._map_algorithm(key_spec.alg) is not None

    def store_secret_supports(self, key_spec):
        """Key storage supported?

        Specifies whether the plugin supports storage of the secret given
        the attributes included in the KeySpec
        """
        return True

    @staticmethod
    def _map_algorithm(algorithm):
        """Map Barbican algorithms to Dogtag plugin algorithms.

        Note that only algorithms supported by Dogtag will be mapped.
        """
        if algorithm is None:
            return None

        if algorithm.lower() == sstore.KeyAlgorithm.AES.lower():
            return key.KeyClient.AES_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.DES.lower():
            return key.KeyClient.DES_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.DESEDE.lower():
            return key.KeyClient.DES3_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.DSA.lower():
            return key.KeyClient.DSA_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.RSA.lower():
            return key.KeyClient.RSA_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.DIFFIE_HELLMAN.lower():
            # may be supported, needs to be tested
            return None
        elif algorithm.lower() == sstore.KeyAlgorithm.EC.lower():
            # asymmetric keys not yet supported
            return None
        else:
            return None

    @staticmethod
    def _store_secret_attributes(meta_dict, secret_dto):
        # store the following attributes for retrieval
        key_spec = secret_dto.key_spec
        if key_spec is not None:
            if key_spec.alg is not None:
                meta_dict[DogtagKRAPlugin.ALG] = key_spec.alg
            if key_spec.bit_length is not None:
                meta_dict[DogtagKRAPlugin.BIT_LENGTH] = key_spec.bit_length
            if key_spec.mode is not None:
                meta_dict[DogtagKRAPlugin.SECRET_MODE] = key_spec.mode

    def _get_passphrase_for_a_private_key(self, secret_type, secret_metadata,
                                          key_spec):
        """Retrieve the passphrase for the private key stored in the KRA."""
        if secret_type is None:
            return None
        if key_spec.alg is None:
            return None

        passphrase = None
        if DogtagKRAPlugin.PASSPHRASE_KEY_ID in secret_metadata:
            if key_spec.alg.upper() == key.KeyClient.RSA_ALGORITHM:
                passphrase = self.keyclient.retrieve_key(
                    secret_metadata.get(DogtagKRAPlugin.PASSPHRASE_KEY_ID)
                ).data
            else:
                if key_spec.alg.upper() == key.KeyClient.DSA_ALGORITHM:
                    raise sstore.SecretGeneralException(
                        u._("DSA keys should not have a passphrase in the"
                            " database, for being used during retrieval.")
                    )
                raise sstore.SecretGeneralException(
                    u._("Secrets of type {secret_type} should not have a "
                        "passphrase in the database, for being used during "
                        "retrieval.").format(secret_type=secret_type)
                )

        # note that Barbican expects the passphrase to be base64 encoded when
        # stored, so we need to decode it.
        if passphrase:
            passphrase = base64.b64decode(passphrase)
        return passphrase

    @staticmethod
    def _get_trans_wrapped_session_key(secret_type, secret_metadata):
        twsk = secret_metadata.get('trans_wrapped_session_key', None)
        if secret_type in [sstore.SecretType.PUBLIC,
                           sstore.SecretType.PRIVATE]:
            if twsk:
                raise DogtagPluginNotSupportedException(
                    u._("Encryption using session key is not supported when "
                        "retrieving a {secret_type} "
                        "key.").format(secret_type=secret_type)
                )

        return twsk


def _catch_request_exception(ca_related_function):
    def _catch_ca_unavailable(self, *args, **kwargs):
        try:
            return ca_related_function(self, *args, **kwargs)
        except request_exceptions.RequestException:
            return cm.ResultDTO(
                cm.CertificateStatus.CA_UNAVAILABLE_FOR_REQUEST)

    return _catch_ca_unavailable


def _catch_enrollment_exceptions(ca_related_function):
    def _catch_enrollment_exception(self, *args, **kwargs):
        try:
            return ca_related_function(self, *args, **kwargs)
        except pki.BadRequestException as e:
            return cm.ResultDTO(
                cm.CertificateStatus.CLIENT_DATA_ISSUE_SEEN,
                status_message=e.message)
        except pki.PKIException as e:
            raise cm.CertificateGeneralException(
                u._("Exception thrown by enroll_cert: {message}").format(
                    message=e.message))

    return _catch_enrollment_exception


def _catch_subca_creation_exceptions(ca_related_function):
    def _catch_subca_exception(self, *args, **kwargs):
        try:
            return ca_related_function(self, *args, **kwargs)
        except pki.BadRequestException as e:
            raise exception.BadSubCACreationRequest(reason=e.message)
        except pki.PKIException as e:
            raise exception.SubCACreationErrors(reason=e.message)
        except request_exceptions.RequestException:
            raise exception.SubCACreationErrors(
                reason="Unable to connect to CA")

    return _catch_subca_exception


def _catch_subca_deletion_exceptions(ca_related_function):
    def _catch_subca_exception(self, *args, **kwargs):
        try:
            return ca_related_function(self, *args, **kwargs)
        except pki.ResourceNotFoundException:
            LOG.warning("Sub-CA already deleted")
            pass
        except pki.PKIException as e:
            raise exception.SubCADeletionErrors(reason=e.message)
        except request_exceptions.RequestException:
            raise exception.SubCACreationErrors(
                reason="Unable to connect to CA")

    return _catch_subca_exception


class DogtagCAPlugin(cm.CertificatePluginBase):
    """Implementation of the cert plugin with Dogtag CA as the backend."""

    # order_metadata fields
    PROFILE_ID = "profile_id"

    # plugin_metadata fields
    REQUEST_ID = "request_id"

    def __init__(self, conf=CONF):
        """Constructor - create the cert clients."""
        connection = create_connection(conf, 'ca')
        self.certclient = pki.cert.CertClient(connection)
        self.simple_cmc_profile = conf.dogtag_plugin.simple_cmc_profile
        self.auto_approved_profiles = conf.dogtag_plugin.auto_approved_profiles

        self.working_dir = conf.dogtag_plugin.plugin_working_dir
        if not os.path.isdir(self.working_dir):
            os.mkdir(self.working_dir)

        self._expiration = None
        self._expiration_delta = conf.dogtag_plugin.ca_expiration_time
        self._expiration_data_path = os.path.join(self.working_dir,
                                                  "expiration_data.txt")

        self._host_aid_path = os.path.join(self.working_dir, "host_aid.txt")
        self._host_aid = None

        if not os.path.isfile(self._expiration_data_path):
            self.expiration = datetime.datetime.utcnow()

        global subcas_available
        subcas_available = self._are_subcas_enabled_on_backend(connection)
        if subcas_available:
            self.authority_client = authority.AuthorityClient(connection)
            if not os.path.isfile(self._host_aid_path):
                self.host_aid = self.get_host_aid()

    @property
    def expiration(self):
        if self._expiration is None:
            try:
                with open(self._expiration_data_path) as expiration_fh:
                    self._expiration = datetime.datetime.strptime(
                        expiration_fh.read(),
                        "%Y-%m-%d %H:%M:%S.%f"
                    )
            except (ValueError, TypeError):
                LOG.warning("Invalid data read from expiration file")
                self.expiration = datetime.utcnow()
        return self._expiration

    @expiration.setter
    def expiration(self, val):
        with open(self._expiration_data_path, 'w') as expiration_fh:
            expiration_fh.write(val.strftime("%Y-%m-%d %H:%M:%S.%f"))
        self._expiration = val

    @property
    def host_aid(self):
        if self._host_aid is None:
            with open(self._host_aid_path) as host_aid_fh:
                self._host_aid = host_aid_fh.read()
        return self._host_aid

    @host_aid.setter
    def host_aid(self, val):
        if val is not None:
            with open(self._host_aid_path, 'w') as host_aid_fh:
                host_aid_fh.write(val)
        self._host_aid = val

    def _are_subcas_enabled_on_backend(self, connection):
        """Check if subca feature is available

        SubCA creation must be supported in both the Dogtag client as well
        as on the back-end server.  Moreover, it must be enabled on the
        backend server.  This method sets the subcas_available global variable.
        :return: True/False
        """
        global subcas_available
        if subcas_available:
            # subcas are supported in the Dogtag client
            try:
                feature_client = feature.FeatureClient(connection)
                authority_feature = feature_client.get_feature("authority")
                if authority_feature.enabled:
                    LOG.info("Sub-CAs are enabled by Dogtag server")
                    return True
                else:
                    LOG.info("Sub-CAs are not enabled by Dogtag server")
            except (request_exceptions.HTTPError,
                    pki.ResourceNotFoundException):
                LOG.info("Sub-CAs are not supported by Dogtag server")
        else:
            LOG.info("Sub-CAs are not supported by Dogtag client")
        return False

    def _get_request_id(self, order_id, plugin_meta, operation):
        request_id = plugin_meta.get(self.REQUEST_ID, None)
        if not request_id:
            raise cm.CertificateGeneralException(
                u._(
                    "{request} not found for {operation} for "
                    "order_id {order_id}"
                ).format(
                    request=self.REQUEST_ID,
                    operation=operation,
                    order_id=order_id
                )
            )
        return request_id

    @_catch_request_exception
    def _get_request(self, request_id):
        try:
            return self.certclient.get_request(request_id)
        except pki.RequestNotFoundException:
            return None

    @_catch_request_exception
    def _get_cert(self, cert_id):
        try:
            return self.certclient.get_cert(cert_id)
        except pki.CertNotFoundException:
            return None

    def get_default_ca_name(self):
        return "Dogtag CA"

    def get_default_signing_cert(self):
        # TODO(alee) Add code to get the signing cert
        return None

    def get_default_intermediates(self):
        # TODO(alee) Add code to get the cert chain
        return None

    def check_certificate_status(self, order_id, order_meta, plugin_meta,
                                 barbican_meta_dto):
        """Check the status of a certificate request.

        :param order_id: ID of the order associated with this request
        :param order_meta: order_metadata associated with this order
        :param plugin_meta: data populated by previous calls for this order,
            in particular the request_id
        :param barbican_meta_dto: additional data needed to process order.
        :return: cm.ResultDTO
        """
        request_id = self._get_request_id(order_id, plugin_meta, "checking")

        request = self._get_request(request_id)
        if not request:
            raise cm.CertificateGeneralException(
                u._(
                    "No request found for request_id {request_id} for "
                    "order {order_id}"
                ).format(
                    request_id=request_id,
                    order_id=order_id
                )
            )

        request_status = request.request_status

        if request_status == pki.cert.CertRequestStatus.REJECTED:
            return cm.ResultDTO(
                cm.CertificateStatus.CLIENT_DATA_ISSUE_SEEN,
                status_message=request.error_message)
        elif request_status == pki.cert.CertRequestStatus.CANCELED:
            return cm.ResultDTO(
                cm.CertificateStatus.REQUEST_CANCELED)
        elif request_status == pki.cert.CertRequestStatus.PENDING:
            return cm.ResultDTO(
                cm.CertificateStatus.WAITING_FOR_CA)
        elif request_status == pki.cert.CertRequestStatus.COMPLETE:
            # get the cert
            cert_id = request.cert_id
            if not cert_id:
                raise cm.CertificateGeneralException(
                    u._(
                        "Request {request_id} reports status_complete, but no "
                        "cert_id has been returned"
                    ).format(
                        request_id=request_id
                    )
                )

            cert = self._get_cert(cert_id)
            if not cert:
                raise cm.CertificateGeneralException(
                    u._("Certificate not found for cert_id: {cert_id}").format(
                        cert_id=cert_id
                    )
                )
            return cm.ResultDTO(
                cm.CertificateStatus.CERTIFICATE_GENERATED,
                certificate=cert.encoded,
                intermediates=cert.pkcs7_cert_chain)
        else:
            raise cm.CertificateGeneralException(
                u._("Invalid request_status returned by CA"))

    @_catch_request_exception
    def issue_certificate_request(self, order_id, order_meta, plugin_meta,
                                  barbican_meta_dto):
        """Issue a certificate request to the Dogtag CA

         Call the relevant certificate issuance function depending on the
         Barbican defined request type in the order_meta.

        :param order_id: ID of the order associated with this request
        :param order_meta: dict containing all the inputs for this request.
               This includes the request_type.
        :param plugin_meta: Used to store data for status check
        :param barbican_meta_dto: additional data needed to process order.
        :return: cm.ResultDTO
        """
        request_type = order_meta.get(
            cm.REQUEST_TYPE,
            cm.CertificateRequestType.CUSTOM_REQUEST)

        jump_table = {
            cm.CertificateRequestType.SIMPLE_CMC_REQUEST:
            self._issue_simple_cmc_request,
            cm.CertificateRequestType.FULL_CMC_REQUEST:
            self._issue_full_cmc_request,
            cm.CertificateRequestType.STORED_KEY_REQUEST:
            self._issue_stored_key_request,
            cm.CertificateRequestType.CUSTOM_REQUEST:
            self._issue_custom_certificate_request
        }

        if request_type not in jump_table:
            raise DogtagPluginNotSupportedException(u._(
                "Dogtag plugin does not support %s request type").format(
                    request_type))

        return jump_table[request_type](order_id, order_meta, plugin_meta,
                                        barbican_meta_dto)

    @_catch_enrollment_exceptions
    def _issue_simple_cmc_request(self, order_id, order_meta, plugin_meta,
                                  barbican_meta_dto):
        """Issue a simple CMC request to the Dogtag CA.

        :param order_id:
        :param order_meta:
        :param plugin_meta:
        :param barbican_meta_dto:
        :return: cm.ResultDTO
        """
        if barbican_meta_dto.generated_csr is not None:
            csr = barbican_meta_dto.generated_csr
        else:
            # we expect the CSR to be base64 encoded PEM
            # Dogtag CA needs it to be unencoded
            csr = base64.b64decode(order_meta.get('request_data'))

        profile_id = order_meta.get('profile', self.simple_cmc_profile)
        inputs = {
            'cert_request_type': 'pkcs10',
            'cert_request': csr
        }

        return self._issue_certificate_request(
            profile_id, inputs, plugin_meta, barbican_meta_dto)

    @_catch_enrollment_exceptions
    def _issue_full_cmc_request(self, order_id, order_meta, plugin_meta,
                                barbican_meta_dto):
        """Issue a full CMC request to the Dogtag CA.

        :param order_id:
        :param order_meta:
        :param plugin_meta:
        :param barbican_meta_dto:
        :return: cm.ResultDTO
        """
        raise DogtagPluginNotSupportedException(u._(
            "Dogtag plugin does not support %s request type").format(
                cm.CertificateRequestType.FULL_CMC_REQUEST))

    @_catch_enrollment_exceptions
    def _issue_stored_key_request(self, order_id, order_meta, plugin_meta,
                                  barbican_meta_dto):
        """Issue a simple CMC request to the Dogtag CA.

        :param order_id:
        :param order_meta:
        :param plugin_meta:
        :param barbican_meta_dto:
        :return: cm.ResultDTO
        """
        return self._issue_simple_cmc_request(
            order_id,
            order_meta,
            plugin_meta,
            barbican_meta_dto)

    @_catch_enrollment_exceptions
    def _issue_custom_certificate_request(self, order_id, order_meta,
                                          plugin_meta, barbican_meta_dto):
        """Issue a custom certificate request to Dogtag CA

        :param order_id: ID of the order associated with this request
        :param order_meta: dict containing all the inputs required for a
            particular profile.  One of these must be the profile_id.
            The exact fields (both optional and mandatory) depend on the
            profile, but they will be exposed to the user in a method to
            expose syntax.  Depending on the profile, only the relevant fields
            will be populated in the request.  All others will be ignored.
        :param plugin_meta: Used to store data for status check.
        :param barbican_meta_dto: Extra data to aid in processing.
        :return: cm.ResultDTO
        """
        profile_id = order_meta.get(self.PROFILE_ID, None)
        if not profile_id:
            return cm.ResultDTO(
                cm.CertificateStatus.CLIENT_DATA_ISSUE_SEEN,
                status_message=u._("No profile_id specified"))

        # we expect the csr to be base64 encoded PEM data.  Dogtag CA expects
        # PEM data though so we need to decode it.
        updated_meta = copy.deepcopy(order_meta)
        if 'cert_request' in updated_meta:
            updated_meta['cert_request'] = base64.b64decode(
                updated_meta['cert_request'])

        return self._issue_certificate_request(
            profile_id, updated_meta, plugin_meta, barbican_meta_dto)

    def _issue_certificate_request(self, profile_id, inputs, plugin_meta,
                                   barbican_meta_dto):
        """Actually send the cert request to the Dogtag CA

        If the profile_id is one of the auto-approved profiles, then use
        the convenience enroll_cert() method to create and approve the request
        using the Barbican agent cert credentials.  If not, then submit the
        request and wait for approval by a CA agent on the Dogtag CA.

        :param profile_id: enrollment profile
        :param inputs: dict of request inputs
        :param plugin_meta: Used to store data for status check.
        :param barbican_meta_dto: Extra data to aid in processing.
        :return: cm.ResultDTO
        """
        ca_id = barbican_meta_dto.plugin_ca_id or self.get_default_ca_name()

        if profile_id in self.auto_approved_profiles:
            if ca_id == self.get_default_ca_name():
                results = self.certclient.enroll_cert(profile_id, inputs)
            else:
                results = self.certclient.enroll_cert(
                    profile_id, inputs, ca_id)
            return self._process_auto_enrollment_results(
                results, plugin_meta, barbican_meta_dto)
        else:
            request = self.certclient.create_enrollment_request(
                profile_id, inputs)
            if ca_id == self.get_default_ca_name():
                results = self.certclient.submit_enrollment_request(request)
            else:
                results = self.certclient.submit_enrollment_request(
                    request, ca_id)
            return self._process_pending_enrollment_results(
                results, plugin_meta, barbican_meta_dto)

    def _process_auto_enrollment_results(self, enrollment_results,
                                         plugin_meta, barbican_meta_dto):
        """Process results received from Dogtag CA for auto-enrollment

        This processes data from enroll_cert, which submits, approves and
        gets the cert issued and returns as a list of CertEnrollment objects.

        :param enrollment_results: list of CertEnrollmentResult objects
        :param plugin_meta: metadata dict for storing plugin specific data
        :param barbican_meta_dto: object containing extra data to help process
               the request
        :return: cm.ResultDTO
        """

        # Although it is possible to create multiple certs in an invocation
        # of enroll_cert, Barbican cannot handle this case.  Assume
        # only once cert and request generated for now.
        enrollment_result = enrollment_results[0]
        request = enrollment_result.request
        if not request:
            raise cm.CertificateGeneralException(
                u._("No request returned in enrollment_results"))

        # store the request_id in the plugin metadata
        plugin_meta[self.REQUEST_ID] = request.request_id

        cert = enrollment_result.cert

        return self._create_dto(request.request_status,
                                request.request_id,
                                request.error_message,
                                cert)

    def _process_pending_enrollment_results(self, results, plugin_meta,
                                            barbican_meta_dto):
        """Process results received from Dogtag CA for pending enrollment

        This method processes data returned by submit_enrollment_request(),
        which creates requests that still need to be approved by an agent.

        :param results: CertRequestInfoCollection object
        :param plugin_meta: metadata dict for storing plugin specific data
        :param barbican_meta_dto: object containing extra data to help process
               the request
        :return: cm.ResultDTO
        """

        # Although it is possible to create multiple certs in an invocation
        # of enroll_cert, Barbican cannot handle this case.  Assume
        # only once cert and request generated for now

        cert_request_info = results.cert_request_info_list[0]
        status = cert_request_info.request_status
        request_id = getattr(cert_request_info, 'request_id', None)
        error_message = getattr(cert_request_info, 'error_message', None)

        # store the request_id in the plugin metadata
        if request_id:
            plugin_meta[self.REQUEST_ID] = request_id

        return self._create_dto(status, request_id, error_message, None)

    def _create_dto(self, request_status, request_id, error_message, cert):
        dto = None
        if request_status == pki.cert.CertRequestStatus.COMPLETE:
            if cert is not None:
                # Barbican is expecting base 64 encoded PEM, so we base64
                # encode below.
                #
                # Currently there is an inconsistency in what Dogtag returns
                # for certificates and intermediates.  For certs, we return
                # PEM, whereas for intermediates, we return headerless PEM.
                # This is being addressed in Dogtag ticket:
                # https://fedorahosted.org/pki/ticket/1374
                #
                # Until this is addressed, simply add the missing headers
                cert_chain = (CERT_HEADER + "\r\n" + cert.pkcs7_cert_chain +
                              CERT_FOOTER)

                dto = cm.ResultDTO(cm.CertificateStatus.CERTIFICATE_GENERATED,
                                   certificate=base64.b64encode(cert.encoded),
                                   intermediates=base64.b64encode(cert_chain))
            else:
                raise cm.CertificateGeneralException(
                    u._("request_id {req_id} returns COMPLETE but no cert "
                        "returned").format(req_id=request_id))

        elif request_status == pki.cert.CertRequestStatus.REJECTED:
            dto = cm.ResultDTO(cm.CertificateStatus.CLIENT_DATA_ISSUE_SEEN,
                               status_message=error_message)
        elif request_status == pki.cert.CertRequestStatus.CANCELED:
            dto = cm.ResultDTO(cm.CertificateStatus.REQUEST_CANCELED)
        elif request_status == pki.cert.CertRequestStatus.PENDING:
            dto = cm.ResultDTO(cm.CertificateStatus.WAITING_FOR_CA)
        else:
            raise cm.CertificateGeneralException(
                u._("Invalid request_status {status} for "
                    "request_id {request_id}").format(
                        status=request_status,
                        request_id=request_id)
            )

        return dto

    def modify_certificate_request(self, order_id, order_meta, plugin_meta,
                                   barbican_meta_dto):
        """Modify a certificate request.

        Once a certificate request is generated, it cannot be modified.
        The only alternative is to cancel the request (if it has not already
        completed) and attempt a fresh enrolment.  That is what will be
        attempted here.
        :param order_id: ID for this order
        :param order_meta: order metadata.  It is assumed that the newly
            modified request data will be present here.
        :param plugin_meta: data stored on behalf of the plugin for further
            operations
        :param barbican_meta_dto: additional data needed to process order.
        :return: ResultDTO:
        """
        result_dto = self.cancel_certificate_request(
            order_id, order_meta, plugin_meta, barbican_meta_dto)

        if result_dto.status == cm.CertificateStatus.REQUEST_CANCELED:
            return self.issue_certificate_request(
                order_id, order_meta, plugin_meta, barbican_meta_dto)
        elif result_dto.status == cm.CertificateStatus.INVALID_OPERATION:
            return cm.ResultDTO(
                cm.CertificateStatus.INVALID_OPERATION,
                status_message=u._(
                    "Modify request: unable to cancel: "
                    "{message}").format(message=result_dto.status_message)
            )
        else:
            # other status (ca_unavailable, client_data_issue)
            # return result from cancel operation
            return result_dto

    @_catch_request_exception
    def cancel_certificate_request(self, order_id, order_meta, plugin_meta,
                                   barbican_meta_dto):
        """Cancel a certificate request.

        :param order_id: ID for the order associated with this request
        :param order_meta: order metadata fdr this request
        :param plugin_meta: data stored by plugin for further processing.
            In particular, the request_id
        :param barbican_meta_dto: additional data needed to process order.
        :return: cm.ResultDTO:
        """
        request_id = self._get_request_id(order_id, plugin_meta, "cancelling")

        try:
            review_response = self.certclient.review_request(request_id)
            self.certclient.cancel_request(request_id, review_response)

            return cm.ResultDTO(cm.CertificateStatus.REQUEST_CANCELED)
        except pki.RequestNotFoundException:
            return cm.ResultDTO(
                cm.CertificateStatus.CLIENT_DATA_ISSUE_SEEN,
                status_message=u._("no request found for this order"))
        except pki.ConflictingOperationException as e:
            return cm.ResultDTO(
                cm.CertificateStatus.INVALID_OPERATION,
                status_message=e.message)

    def supports(self, certificate_spec):
        if cm.CA_TYPE in certificate_spec:
            return certificate_spec[cm.CA_TYPE] == cm.CA_PLUGIN_TYPE_DOGTAG

        if cm.CA_PLUGIN_TYPE_SYMANTEC in certificate_spec:
            # TODO(alee-3) Handle case where SKI is provided
            pass

        return True

    def supported_request_types(self):
        """Returns the request_types supported by this plugin.

        :returns: a list of the Barbican-core defined request_types
                  supported by this plugin.
        """
        return [cm.CertificateRequestType.SIMPLE_CMC_REQUEST,
                cm.CertificateRequestType.STORED_KEY_REQUEST,
                cm.CertificateRequestType.CUSTOM_REQUEST]

    def supports_create_ca(self):
        """Returns if this plugin and the backend CA supports subCAs

        :return: True/False
        """
        return subcas_available

    @_catch_subca_creation_exceptions
    def create_ca(self, ca_create_dto):
        """Creates a subordinate CA upon request

        :param ca_create_dto:
            Data transfer object :class:`CACreateDTO` containing data
            required to generate a subordinate CA.  This data includes
            the subject DN of the new CA signing certificate, a name for
            the new CA and a reference to the CA that will issue the new
            subordinate CA's signing certificate,

        :return: ca_info:
            Dictionary containing the data needed to create a
            models.CertificateAuthority object
        """
        if not subcas_available:
            raise exception.SubCAsNotSupported(
                "Subordinate CAs are not supported by this Dogtag CA")

        parent_ca_id = self._get_correct_ca_id(ca_create_dto.parent_ca_id)
        ca_data = authority.AuthorityData(
            dn=ca_create_dto.subject_dn,
            parent_aid=parent_ca_id,
            description=ca_create_dto.name)

        new_ca_data = self.authority_client.create_ca(ca_data)

        cert = self.authority_client.get_cert(new_ca_data.aid, "PEM")
        chain = self.authority_client.get_chain(new_ca_data.aid, "PEM")

        return {
            cm.INFO_NAME: new_ca_data.description,
            cm.INFO_CA_SIGNING_CERT: cert,
            cm.INFO_EXPIRATION: self.expiration.isoformat(),
            cm.INFO_INTERMEDIATES: chain,
            cm.PLUGIN_CA_ID: new_ca_data.aid
        }

    def _get_correct_ca_id(self, plugin_ca_id):
        """Returns the correct authority id

        When the Dogtag plugin updates its CA list, any subcas will
        have a plugin_ca_id that matches the authority_id (aid) as
        returned from the backend CA.

        For migration purposes, though, ie. migrating from a non-subca
        environment to a subca one, we want the host CA to keep the same
        plugin_ca_id (which is the default_ca_name) so that no disruption
        occurs.  Therefore, we need to store the host CA's authority ID
        (in get_ca_info) and return it here instead.
        """
        if plugin_ca_id == self.get_default_ca_name():
            return self.host_aid
        else:
            return plugin_ca_id

    @_catch_subca_deletion_exceptions
    def delete_ca(self, ca_id):
        """Deletes a subordinate CA

        :param ca_id: id for the CA as specified by the plugin
        :return: None
        """
        if not subcas_available:
            raise exception.SubCAsNotSupported(
                "Subordinate CAs are not supported by this Dogtag CA")

        # ca must be disabled first
        self.authority_client.disable_ca(ca_id)
        self.authority_client.delete_ca(ca_id)

    def get_ca_info(self):
        if not subcas_available:
            return super(DogtagCAPlugin, self).get_ca_info()

        self.expiration = (datetime.datetime.utcnow() + datetime.timedelta(
            days=int(self._expiration_delta)))

        ret = {}
        cas = self.authority_client.list_cas()
        for ca_data in cas.ca_list:
            if not ca_data.enabled:
                continue

            cert = self.authority_client.get_cert(ca_data.aid, "PEM")
            chain = self.authority_client.get_chain(ca_data.aid, "PEM")
            ca_info = {
                cm.INFO_NAME: ca_data.description,
                cm.INFO_CA_SIGNING_CERT: cert,
                cm.INFO_INTERMEDIATES: chain,
                cm.INFO_EXPIRATION: self.expiration.isoformat()
            }

            # handle the migration case.  The top level CA should continue
            # to work as before

            if ca_data.is_host_authority:
                ret[self.get_default_ca_name()] = ca_info
                self.host_aid = ca_data.aid
            else:
                ret[ca_data.aid] = ca_info

        return ret

    def get_host_aid(self):
        cas = self.authority_client.list_cas()
        for ca_data in cas.ca_list:
            if ca_data.is_host_authority:
                return ca_data.aid
        return None



# Copyright (c) 2018 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import six

from castellan.common.objects import opaque_data
from castellan import key_manager
from oslo_context import context
from oslo_log import log

from barbican.plugin.interface import secret_store as ss

LOG = log.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class CastellanSecretStore(ss.SecretStoreBase):

    KEY_ID = "key_id"
    ALG = "alg"
    BIT_LENGTH = "bit_length"

    def _set_params(self, conf):
        self.key_manager = key_manager.API(conf)
        self.context = context.get_current()

    @abc.abstractmethod
    def get_conf(self, conf):
        """Get plugin configuration

        This method is supposed to be implemented by the relevant
        subclass.  This method reads in the config for the plugin
        in barbican.conf -- which should look like the way other
        barbican plugins are configured, and convert them to the
        proper oslo.config object to be passed to the keymanager
        API. (keymanager.API(conf)

        @returns oslo.config object
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def get_plugin_name(self):
        """Get plugin name

        This method is implemented by the subclass.
        Note that this name must be unique across the deployment.
        """
        raise NotImplementedError  # pragma: no cover

    def get_secret(self, secret_type, secret_metadata):
        secret_ref = secret_metadata[CastellanSecretStore.KEY_ID]
        try:
            secret = self.key_manager.get(
                self.context,
                secret_ref)

            return ss.SecretDTO(secret_type, secret.get_encoded(),
                                ss.KeySpec(), secret_metadata['content_type'])
        except Exception as e:
            LOG.exception("Error retrieving secret {}: {}".format(
                secret_ref, six.text_type(e)))
            raise ss.SecretGeneralException(e)

    def store_secret(self, secret_dto):
        if not self.store_secret_supports(secret_dto.key_spec):
            raise ss.SecretAlgorithmNotSupportedException(
                secret_dto.key_spec.alg)

        try:
            secret_ref = self.key_manager.store(
                self.context,
                opaque_data.OpaqueData(secret_dto.secret)
            )
            return {CastellanSecretStore.KEY_ID: secret_ref}
        except Exception as e:
            LOG.exception("Error storing secret: {}".format(
                six.text_type(e)))
            raise ss.SecretGeneralException(e)

    def delete_secret(self, secret_metadata):
        secret_ref = secret_metadata[CastellanSecretStore.KEY_ID]
        try:
            self.key_manager.delete(
                self.context,
                secret_ref)
        except KeyError:
            LOG.warning("Attempting to delete a non-existent secret {}".format(
                secret_ref))
        except Exception as e:
            LOG.exception("Error deleting secret: {}".format(
                six.text_type(e)))
            raise ss.SecretGeneralException(e)

    def generate_symmetric_key(self, key_spec):
        if not self.generate_supports(key_spec):
            raise ss.SecretAlgorithmNotSupportedException(
                key_spec.alg)
        try:
            secret_ref = self.key_manager.create_key(
                self.context,
                key_spec.alg,
                key_spec.bit_length
            )
            return {CastellanSecretStore.KEY_ID: secret_ref}
        except Exception as e:
            LOG.exception("Error generating symmetric key: {}".format(
                six.text_type(e)))
            raise ss.SecretGeneralException(e)

    def generate_asymmetric_key(self, key_spec):
        if not self.generate_supports(key_spec):
            raise ss.SecretAlgorithmNotSupportedException(
                key_spec.alg)

        if key_spec.passphrase:
            raise ss.GeneratePassphraseNotSupportedException()

        try:
            private_ref, public_ref = self.key_manager.create_key_pair(
                self.context,
                key_spec.alg,
                key_spec.bit_length
            )

            private_key_metadata = {
                CastellanSecretStore.ALG: key_spec.alg,
                CastellanSecretStore.BIT_LENGTH: key_spec.bit_length,
                CastellanSecretStore.KEY_ID: private_ref
            }

            public_key_metadata = {
                CastellanSecretStore.ALG: key_spec.alg,
                CastellanSecretStore.BIT_LENGTH: key_spec.bit_length,
                CastellanSecretStore.KEY_ID: public_ref
            }

            return ss.AsymmetricKeyMetadataDTO(
                private_key_metadata,
                public_key_metadata,
                None
            )
        except Exception as e:
            LOG.exception("Error generating asymmetric key: {}".format(
                six.text_type(e)))
            raise ss.SecretGeneralException(e)

    @abc.abstractmethod
    def store_secret_supports(self, key_spec):
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def generate_supports(self, key_spec):
        raise NotImplementedError  # pragma: no cover
